# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-module-docstring
# pylint: disable=protected-access
# pylint: disable=redefined-outer-name

import pytest

from speciesnet.taxonomy_utils import get_ancestor_at_level
from speciesnet.taxonomy_utils import get_full_class_string

# fmt: off
# pylint: disable=line-too-long

BLANK = "f1856211-cfb7-4a5b-9158-c0f72fd09ee6;;;;;;blank"
BLANK_FC = ";;;;"
HUMAN = "990ae9dd-7a59-4344-afcb-1b7b21368000;mammalia;primates;hominidae;homo;sapiens;human"
HUMAN_FC = "mammalia;primates;hominidae;homo;sapiens"
VEHICLE = "e2895ed5-780b-48f6-8a11-9e27cb594511;;;;;;vehicle"
VEHICLE_FC = ";;;;"

LION = "ddf59264-185a-4d35-b647-2785792bdf54;mammalia;carnivora;felidae;panthera;leo;lion"
LION_FC = "mammalia;carnivora;felidae;panthera;leo"
PANTHERA_GENUS = "fbb23d07-6677-43db-b650-f99ac452c50f;mammalia;carnivora;felidae;panthera;;panthera species"
PANTHERA_GENUS_FC = "mammalia;carnivora;felidae;panthera;"
FELIDAE_FAMILY = "df8514b0-10a5-411f-8ed6-0f415e8153a3;mammalia;carnivora;felidae;;;cat family"
FELIDAE_FAMILY_FC = "mammalia;carnivora;felidae;;"
CARNIVORA_ORDER = "eeeb5d26-2a47-4d01-a3de-10b33ec0aee4;mammalia;carnivora;;;;carnivorous mammal"
CARNIVORA_ORDER_FC = "mammalia;carnivora;;;"
MAMMALIA_CLASS = "f2d233e3-80e3-433d-9687-e29ecc7a467a;mammalia;;;;;mammal"
MAMMALIA_CLASS_FC = "mammalia;;;;"
ANIMAL_KINGDOM = "1f689929-883d-4dae-958c-3d57ab5b6c16;;;;;;animal"
ANIMAL_KINGDOM_FC = ";;;;"

BROWN_BEAR = "330bb1e9-84d6-4e41-afa9-938aee17ea29;mammalia;carnivora;ursidae;ursus;arctos;brown bear"
BROWN_BEAR_FC = "mammalia;carnivora;ursidae;ursus;arctos"
POLAR_BEAR = "e7f83bf6-df2c-4ce0-97fc-2f233df23ec4;mammalia;carnivora;ursidae;ursus;maritimus;polar bear"
POLAR_BEAR_FC = "mammalia;carnivora;ursidae;ursus;maritimus"
GIANT_PANDA = "85662682-67c1-4ecb-ba05-ba12e2df6b65;mammalia;carnivora;ursidae;ailuropoda;melanoleuca;giant panda"
GIANT_PANDA_FC = "mammalia;carnivora;ursidae;ailuropoda;melanoleuca"
URSUS_GENUS = "5a0f5e3f-c634-4b86-910a-b105cb526a24;mammalia;carnivora;ursidae;ursus;;ursus species"
URSUS_GENUS_FC = "mammalia;carnivora;ursidae;ursus;"
URSIDAE_FAMILY = "ec1a70f4-41c0-4aba-9150-292fb2b7a324;mammalia;carnivora;ursidae;;;bear family"
URSIDAE_FAMILY_FC = "mammalia;carnivora;ursidae;;"

PUMA = "9c564562-9429-405c-8529-04cff7752282;mammalia;carnivora;felidae;puma;concolor;puma"
PUMA_FC = "mammalia;carnivora;felidae;puma;concolor"
SAND_CAT = "e588253d-d61d-4149-a96c-8c245927a80f;mammalia;carnivora;felidae;felis;margarita;sand cat"
SAND_CAT_FC = "mammalia;carnivora;felidae;felis;margarita"

# pylint: enable=line-too-long
# fmt: on


@pytest.fixture
def taxonomy_map():
    return {
        BLANK_FC: BLANK,
        HUMAN_FC: HUMAN,
        VEHICLE_FC: VEHICLE,
        LION_FC: LION,
        PANTHERA_GENUS_FC: PANTHERA_GENUS,
        FELIDAE_FAMILY_FC: FELIDAE_FAMILY,
        CARNIVORA_ORDER_FC: CARNIVORA_ORDER,
        MAMMALIA_CLASS_FC: MAMMALIA_CLASS,
        ANIMAL_KINGDOM_FC: ANIMAL_KINGDOM,
        BROWN_BEAR_FC: BROWN_BEAR,
        POLAR_BEAR_FC: POLAR_BEAR,
        GIANT_PANDA_FC: GIANT_PANDA,
        URSUS_GENUS_FC: URSUS_GENUS,
        URSIDAE_FAMILY_FC: URSIDAE_FAMILY,
    }


class TestTaxonomyUtils:
    """Tests for the taxonomy utility functions."""

    def test_get_ancestor_at_level(self, taxonomy_map) -> None:

        # Test all ancestors of LION.
        assert get_ancestor_at_level(LION, "species", taxonomy_map) == LION
        assert get_ancestor_at_level(LION, "genus", taxonomy_map) == PANTHERA_GENUS
        assert get_ancestor_at_level(LION, "family", taxonomy_map) == FELIDAE_FAMILY
        assert get_ancestor_at_level(LION, "order", taxonomy_map) == CARNIVORA_ORDER
        assert get_ancestor_at_level(LION, "class", taxonomy_map) == MAMMALIA_CLASS
        assert get_ancestor_at_level(LION, "kingdom", taxonomy_map) == ANIMAL_KINGDOM

        # Test all ancestors of PANTHERA_GENUS.
        assert get_ancestor_at_level(PANTHERA_GENUS, "species", taxonomy_map) is None
        assert (
            get_ancestor_at_level(PANTHERA_GENUS, "genus", taxonomy_map)
            == PANTHERA_GENUS
        )
        assert (
            get_ancestor_at_level(PANTHERA_GENUS, "family", taxonomy_map)
            == FELIDAE_FAMILY
        )
        assert (
            get_ancestor_at_level(PANTHERA_GENUS, "order", taxonomy_map)
            == CARNIVORA_ORDER
        )
        assert (
            get_ancestor_at_level(PANTHERA_GENUS, "class", taxonomy_map)
            == MAMMALIA_CLASS
        )
        assert (
            get_ancestor_at_level(PANTHERA_GENUS, "kingdom", taxonomy_map)
            == ANIMAL_KINGDOM
        )

        # Test all ancestors of FELIDAE_FAMILY.
        assert get_ancestor_at_level(FELIDAE_FAMILY, "species", taxonomy_map) is None
        assert get_ancestor_at_level(FELIDAE_FAMILY, "genus", taxonomy_map) is None
        assert (
            get_ancestor_at_level(FELIDAE_FAMILY, "family", taxonomy_map)
            == FELIDAE_FAMILY
        )
        assert (
            get_ancestor_at_level(FELIDAE_FAMILY, "order", taxonomy_map)
            == CARNIVORA_ORDER
        )
        assert (
            get_ancestor_at_level(FELIDAE_FAMILY, "class", taxonomy_map)
            == MAMMALIA_CLASS
        )
        assert (
            get_ancestor_at_level(FELIDAE_FAMILY, "kingdom", taxonomy_map)
            == ANIMAL_KINGDOM
        )

        # Test all ancestors of CARNIVORA_ORDER.
        assert get_ancestor_at_level(CARNIVORA_ORDER, "species", taxonomy_map) is None
        assert get_ancestor_at_level(CARNIVORA_ORDER, "genus", taxonomy_map) is None
        assert get_ancestor_at_level(CARNIVORA_ORDER, "family", taxonomy_map) is None
        assert (
            get_ancestor_at_level(CARNIVORA_ORDER, "order", taxonomy_map)
            == CARNIVORA_ORDER
        )
        assert (
            get_ancestor_at_level(CARNIVORA_ORDER, "class", taxonomy_map)
            == MAMMALIA_CLASS
        )
        assert (
            get_ancestor_at_level(CARNIVORA_ORDER, "kingdom", taxonomy_map)
            == ANIMAL_KINGDOM
        )

        # Test all ancestors of MAMMALIA_CLASS.
        assert get_ancestor_at_level(MAMMALIA_CLASS, "species", taxonomy_map) is None
        assert get_ancestor_at_level(MAMMALIA_CLASS, "genus", taxonomy_map) is None
        assert get_ancestor_at_level(MAMMALIA_CLASS, "family", taxonomy_map) is None
        assert get_ancestor_at_level(MAMMALIA_CLASS, "order", taxonomy_map) is None
        assert (
            get_ancestor_at_level(MAMMALIA_CLASS, "class", taxonomy_map)
            == MAMMALIA_CLASS
        )
        assert (
            get_ancestor_at_level(MAMMALIA_CLASS, "kingdom", taxonomy_map)
            == ANIMAL_KINGDOM
        )

        # Test all ancestors of ANIMAL_KINGDOM.
        assert get_ancestor_at_level(ANIMAL_KINGDOM, "species", taxonomy_map) is None
        assert get_ancestor_at_level(ANIMAL_KINGDOM, "genus", taxonomy_map) is None
        assert get_ancestor_at_level(ANIMAL_KINGDOM, "family", taxonomy_map) is None
        assert get_ancestor_at_level(ANIMAL_KINGDOM, "order", taxonomy_map) is None
        assert get_ancestor_at_level(ANIMAL_KINGDOM, "class", taxonomy_map) is None
        assert (
            get_ancestor_at_level(ANIMAL_KINGDOM, "kingdom", taxonomy_map)
            == ANIMAL_KINGDOM
        )

        # Test all ancestors of BLANK.
        assert get_ancestor_at_level(BLANK, "species", taxonomy_map) is None
        assert get_ancestor_at_level(BLANK, "genus", taxonomy_map) is None
        assert get_ancestor_at_level(BLANK, "family", taxonomy_map) is None
        assert get_ancestor_at_level(BLANK, "order", taxonomy_map) is None
        assert get_ancestor_at_level(BLANK, "class", taxonomy_map) is None
        assert get_ancestor_at_level(BLANK, "kingdom", taxonomy_map) is None

        # Test all ancestors of HUMAN, when its genus, family and order are missing from
        # the mock taxonomy mapping.
        assert get_ancestor_at_level(HUMAN, "species", taxonomy_map) == HUMAN
        assert get_ancestor_at_level(HUMAN, "genus", taxonomy_map) is None
        assert get_ancestor_at_level(HUMAN, "family", taxonomy_map) is None
        assert get_ancestor_at_level(HUMAN, "order", taxonomy_map) is None
        assert get_ancestor_at_level(HUMAN, "class", taxonomy_map) == MAMMALIA_CLASS
        assert get_ancestor_at_level(HUMAN, "kingdom", taxonomy_map) == ANIMAL_KINGDOM

        # Test all ancestors of VEHICLE.
        assert get_ancestor_at_level(VEHICLE, "species", taxonomy_map) is None
        assert get_ancestor_at_level(VEHICLE, "genus", taxonomy_map) is None
        assert get_ancestor_at_level(VEHICLE, "family", taxonomy_map) is None
        assert get_ancestor_at_level(VEHICLE, "order", taxonomy_map) is None
        assert get_ancestor_at_level(VEHICLE, "class", taxonomy_map) is None
        assert get_ancestor_at_level(VEHICLE, "kingdom", taxonomy_map) is None

        # Test all ancestors of an unseen species.
        unseen_species = "uuid;class;order;family;genus;species;common_name"
        assert get_ancestor_at_level(unseen_species, "species", taxonomy_map) is None
        assert get_ancestor_at_level(unseen_species, "genus", taxonomy_map) is None
        assert get_ancestor_at_level(unseen_species, "family", taxonomy_map) is None
        assert get_ancestor_at_level(unseen_species, "order", taxonomy_map) is None
        assert get_ancestor_at_level(unseen_species, "class", taxonomy_map) is None
        assert (
            get_ancestor_at_level(unseen_species, "kingdom", taxonomy_map)
            == ANIMAL_KINGDOM
        )

        # Test errors due to invalid labels.
        with pytest.raises(ValueError):
            invalid_label = "uuid;class;order;family;genus;species"
            get_ancestor_at_level(invalid_label, "kingdom", taxonomy_map)
        with pytest.raises(ValueError):
            invalid_label = "uuid;class;order;family;genus;species;common_name;extra"
            get_ancestor_at_level(invalid_label, "kingdom", taxonomy_map)

    def test_get_full_class_string(self) -> None:

        # Test BLANK/HUMAN/VEHICLE.
        assert get_full_class_string(BLANK) == BLANK_FC
        assert get_full_class_string(HUMAN) == HUMAN_FC
        assert get_full_class_string(VEHICLE) == VEHICLE_FC

        # Test valid labels at different taxonomy levels.
        assert get_full_class_string(LION) == LION_FC
        assert get_full_class_string(PANTHERA_GENUS) == PANTHERA_GENUS_FC
        assert get_full_class_string(FELIDAE_FAMILY) == FELIDAE_FAMILY_FC
        assert get_full_class_string(CARNIVORA_ORDER) == CARNIVORA_ORDER_FC
        assert get_full_class_string(MAMMALIA_CLASS) == MAMMALIA_CLASS_FC
        assert get_full_class_string(ANIMAL_KINGDOM) == ANIMAL_KINGDOM_FC

        # Test errors due to invalid labels.
        with pytest.raises(ValueError):
            invalid_label = "uuid;class;order;family;genus;species"
            get_full_class_string(invalid_label)
        with pytest.raises(ValueError):
            invalid_label = "uuid;class;order;family;genus;species;common_name;extra"
            get_full_class_string(invalid_label)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-module-docstring

import json
from pathlib import Path
from typing import Generator

import pytest

from speciesnet.utils import file_exists
from speciesnet.utils import load_partial_predictions
from speciesnet.utils import load_rgb_image
from speciesnet.utils import ModelInfo
from speciesnet.utils import prepare_instances_dict
from speciesnet.utils import save_predictions

# fmt: off
# pylint: disable=line-too-long

AZ_VIA_HTTPS_TEST_DIR = "https://lilawildlife.blob.core.windows.net/lila-wildlife/caltech-unzipped"
AZ_VIA_HTTPS_TEST_IMG = "https://lilawildlife.blob.core.windows.net/lila-wildlife/caltech-unzipped/cct_images/59420eea-23d2-11e8-a6a3-ec086b02610b.jpg"
GS_TEST_DIR = "gs://public-datasets-lila/caltech-unzipped"
GS_TEST_IMG = "gs://public-datasets-lila/caltech-unzipped/cct_images/59420eea-23d2-11e8-a6a3-ec086b02610b.jpg"
S3_TEST_DIR = "s3://us-west-2.opendata.source.coop/agentmorris/lila-wildlife/caltech-unzipped"
S3_TEST_IMG = "s3://us-west-2.opendata.source.coop/agentmorris/lila-wildlife/caltech-unzipped/cct_images/59420eea-23d2-11e8-a6a3-ec086b02610b.jpg"

# pylint: enable=line-too-long
# fmt: on


class TestModelInfo:
    """Tests for the ModelInfo dataclass."""

    def test_model_type(self, model_name: str) -> None:
        model_info = ModelInfo(model_name)
        assert model_info.type_ in {"always_crop", "full_image"}


class TestPrepareInstancesDict:
    """Tests for the standardization of input formats."""

    @pytest.fixture
    def filepaths_as_strings(self) -> list[str]:
        return [
            "test_data/african_elephants.jpg",
            "test_data/african_elephants_bw.jpg",
            "test_data/african_elephants_cmyk.jpg",
            "test_data/african_elephants_truncated_file.jpg",
            "test_data/african_elephants_with_exif_orientation.jpg",
            "test_data/american_black_bear.jpg",
            "test_data/blank.jpg",
            "test_data/blank2.jpg",
            "test_data/blank3.jpg",
            "test_data/domestic_cattle.jpg",
            "test_data/domestic_dog.jpg",
            "test_data/human.jpg",
            "test_data/ocelot.jpg",
            "test_data/vehicle.jpg",
        ]

    @pytest.fixture
    def filepaths_as_paths(self, filepaths_as_strings) -> list[Path]:
        return [Path(f) for f in filepaths_as_strings]

    @pytest.fixture
    def folders_as_strings(self) -> list[str]:
        return [
            "test_data",
        ]

    @pytest.fixture
    def folders_as_paths(self, folders_as_strings) -> list[Path]:
        return [Path(f) for f in folders_as_strings]

    @pytest.fixture
    def instances_dict(self, filepaths_as_strings) -> dict:
        return {"instances": [{"filepath": f} for f in filepaths_as_strings]}

    @pytest.fixture
    def instances_dict_portugal(self, filepaths_as_strings) -> dict:
        return {
            "instances": [
                {
                    "filepath": f,
                    "country": "PRT",
                }
                for f in filepaths_as_strings
            ]
        }

    @pytest.fixture
    def instances_dict_portugal_porto(self, filepaths_as_strings) -> dict:
        return {
            "instances": [
                {
                    "filepath": f,
                    "country": "PRT",
                    "admin1_region": "13",
                }
                for f in filepaths_as_strings
            ]
        }

    def test_invalid_inputs(
        self, instances_dict, filepaths_as_strings, folders_as_paths
    ) -> None:
        with pytest.raises(ValueError):
            prepare_instances_dict()
        with pytest.raises(ValueError):
            prepare_instances_dict(
                instances_dict=instances_dict, filepaths=filepaths_as_strings
            )
        with pytest.raises(ValueError):
            prepare_instances_dict(
                instances_dict=instances_dict, folders_txt="test_data/local_folders.txt"
            )
        with pytest.raises(ValueError):
            prepare_instances_dict(
                instances_dict=instances_dict,
                filepaths_txt="test_data/local_filepaths.txt",
                folders=folders_as_paths,
            )

    def test_instances_conversion(self, instances_dict) -> None:
        assert prepare_instances_dict(instances_dict=instances_dict) == instances_dict
        assert (
            prepare_instances_dict(instances_json="test_data/local_instances.json")
            == instances_dict
        )
        assert (
            prepare_instances_dict(
                instances_json=Path("test_data/local_instances.json")
            )
            == instances_dict
        )

    def test_filepaths_conversion(
        self, instances_dict, filepaths_as_strings, filepaths_as_paths
    ) -> None:
        assert prepare_instances_dict(filepaths=filepaths_as_strings) == instances_dict
        assert prepare_instances_dict(filepaths=filepaths_as_paths) == instances_dict
        assert (
            prepare_instances_dict(filepaths_txt="test_data/local_filepaths.txt")
            == instances_dict
        )
        assert (
            prepare_instances_dict(filepaths_txt=Path("test_data/local_filepaths.txt"))
            == instances_dict
        )

    def test_folders_conversion(
        self, instances_dict, folders_as_strings, folders_as_paths
    ) -> None:
        assert prepare_instances_dict(folders=folders_as_strings) == instances_dict
        assert prepare_instances_dict(folders=folders_as_paths) == instances_dict
        assert (
            prepare_instances_dict(folders_txt="test_data/local_folders.txt")
            == instances_dict
        )
        assert (
            prepare_instances_dict(folders_txt=Path("test_data/local_folders.txt"))
            == instances_dict
        )

    def test_country_overwrites(self, instances_dict, instances_dict_portugal) -> None:
        assert (
            prepare_instances_dict(instances_dict=instances_dict, country="PRT")
            == instances_dict_portugal
        )

    def test_admin1_region_overwrites(
        self, instances_dict, instances_dict_portugal_porto
    ) -> None:
        assert (
            prepare_instances_dict(instances_dict=instances_dict, admin1_region="13")
            == instances_dict
        )
        assert (
            prepare_instances_dict(
                instances_dict=instances_dict, country="PRT", admin1_region="13"
            )
            == instances_dict_portugal_porto
        )


class TestFileExists:
    """Tests for the existence of files."""

    def test_local_file(self) -> None:
        assert file_exists("test_data/blank.jpg")
        assert not file_exists("test_data/missing.jpg")

    def test_http_file(self) -> None:
        assert file_exists("http://picsum.photos/200/300")
        assert not file_exists("http://picsum.photos/missing.jpg")

    def test_https_file(self) -> None:
        assert file_exists("https://picsum.photos/300/400")
        assert not file_exists("https://picsum.photos/missing.jpg")

    @pytest.mark.az
    def test_az_via_https_file(self) -> None:
        # We can't test the `az://` scheme directly because it's a pseudo-URL scheme,
        # not officially supported by Azure. However, we still support it via
        # `cloudpathlib` and `az://` URLs should work without issues when the user is
        # properly authenticated:
        # https://github.com/drivendataorg/cloudpathlib/issues/157
        assert file_exists(AZ_VIA_HTTPS_TEST_IMG)
        assert not file_exists(AZ_VIA_HTTPS_TEST_DIR + "/missing.jpg")

    @pytest.mark.gs
    def test_gs_file(self) -> None:
        assert file_exists(GS_TEST_IMG)
        assert not file_exists("gs://missing.jpg")
        assert not file_exists(GS_TEST_DIR + "/missing.jpg")

    @pytest.mark.s3
    def test_s3_file(self) -> None:
        assert file_exists(S3_TEST_IMG)
        assert not file_exists("s3://missing.jpg")
        assert not file_exists(S3_TEST_DIR + "/missing.jpg")


class TestLoadRGBImage:
    """Tests for the image loading."""

    def test_valid_image(self) -> None:
        img = load_rgb_image("test_data/african_elephants.jpg")
        assert img
        assert img.size == (2048, 1536)
        assert img.mode == "RGB"

    def test_missing_image(self) -> None:
        img = load_rgb_image("test_data/missing.jpg")
        assert img is None

    def test_truncated_image(self) -> None:
        img = load_rgb_image("test_data/african_elephants_truncated_file.jpg")
        assert img
        assert img.size == (2048, 1536)
        assert img.mode == "RGB"

    def test_exif_transpose(self) -> None:
        img = load_rgb_image("test_data/african_elephants_with_exif_orientation.jpg")
        assert img
        assert img.size == (2048, 1536)
        assert img.mode == "RGB"

    def test_cmyk_image(self) -> None:
        img = load_rgb_image("test_data/african_elephants_cmyk.jpg")
        assert img
        assert img.size == (2048, 1536)
        assert img.mode == "RGB"

    def test_bw_image(self) -> None:
        img = load_rgb_image("test_data/african_elephants_bw.jpg")
        assert img
        assert img.size == (2048, 1536)
        assert img.mode == "RGB"

    def test_http_image(self) -> None:
        img = load_rgb_image("http://picsum.photos/200/300")
        assert img
        assert img.size == (200, 300)
        assert img.mode == "RGB"

    def test_https_image(self) -> None:
        img = load_rgb_image("https://picsum.photos/300/400")
        assert img
        assert img.size == (300, 400)
        assert img.mode == "RGB"

    @pytest.mark.az
    def test_az_via_https_image(self) -> None:
        # We can't test the `az://` scheme directly because it's a pseudo-URL scheme,
        # not officially supported by Azure. However, we still support it via
        # `cloudpathlib` and `az://` URLs should work without issues when the user is
        # properly authenticated:
        # https://github.com/drivendataorg/cloudpathlib/issues/157
        img = load_rgb_image(AZ_VIA_HTTPS_TEST_IMG)
        assert img
        assert img.size == (2048, 1494)
        assert img.mode == "RGB"

    @pytest.mark.gs
    def test_gs_image(self) -> None:
        img = load_rgb_image(GS_TEST_IMG)
        assert img
        assert img.size == (2048, 1494)
        assert img.mode == "RGB"

    @pytest.mark.s3
    def test_s3_image(self) -> None:
        img = load_rgb_image(S3_TEST_IMG)
        assert img
        assert img.size == (2048, 1494)
        assert img.mode == "RGB"


class TestLoadPartialPredictions:
    """Tests for the loading of partial predictions."""

    def test_no_loading(self, tmp_path) -> None:
        valid_instances = [
            {"filepath": "a.jpg"},
            {"filepath": "b.jpg"},
            {"filepath": "c.jpg"},
        ]
        assert load_partial_predictions(None, valid_instances) == ({}, valid_instances)
        assert load_partial_predictions(tmp_path, valid_instances) == (
            {},
            valid_instances,
        )

    @pytest.fixture
    def valid_predictions_json(self, tmp_path) -> Generator[Path, None, None]:
        json_content = {
            "predictions": [
                {
                    "filepath": "a.jpg",
                    "failures": ["CLASSIFIER", "DETECTOR"],
                    "country": "COUNTRY_A",
                    "model_version": "0.0.0",
                },
                {
                    "filepath": "b.jpg",
                    "failures": ["CLASSIFIER"],
                    "country": "COUNTRY_B",
                    "detections": [
                        {
                            "category": "1",
                            "label": "animal",
                            "conf": 0.5,
                            "bbox": [0.0, 0.1, 0.2, 0.3],
                        }
                    ],
                    "model_version": "0.0.0",
                },
                {
                    "filepath": "c.jpg",
                    "failures": ["DETECTOR"],
                    "country": "COUNTRY_C",
                    "classifications": {
                        "classes": ["X", "Y", "Z"],
                        "scores": [0.5, 0.3, 0.2],
                    },
                    "model_version": "0.0.0",
                },
                {
                    "filepath": "d.jpg",
                    "failures": ["GEOLOCATION"],
                    "classifications": {
                        "classes": ["R", "S", "T"],
                        "scores": [0.7, 0.2, 0.1],
                    },
                    "detections": [
                        {
                            "category": "2",
                            "label": "human",
                            "conf": 0.7,
                            "bbox": [0.1, 0.2, 0.3, 0.4],
                        }
                    ],
                    "prediction": "R",
                    "prediction_score": 0.7,
                    "prediction_source": "test",
                    "model_version": "0.0.0",
                },
                {
                    "filepath": "e.jpg",
                    "country": "COUNTRY_E",
                    "classifications": {
                        "classes": ["K", "L", "M"],
                        "scores": [0.9, 0.1, 0.0],
                    },
                    "detections": [
                        {
                            "category": "2",
                            "label": "human",
                            "conf": 0.7,
                            "bbox": [0.1, 0.2, 0.3, 0.4],
                        }
                    ],
                    "prediction": "K",
                    "prediction_score": 0.9,
                    "prediction_source": "test",
                    "model_version": "0.0.0",
                },
                {
                    "filepath": "f.jpg",
                    "classifications": {
                        "classes": ["XYZ"],
                        "scores": [0.8],
                    },
                    "detections": [],
                    "prediction": "XYZ",
                    "prediction_score": 0.4,
                    "prediction_source": "test",
                    "model_version": "0.0.0",
                },
            ]
        }

        filepath = tmp_path / "valid_predictions.json"
        with open(filepath, mode="w", encoding="utf-8") as fp:
            json.dump(json_content, fp)
        yield filepath
        filepath.unlink()

    def test_successful_loading(self, valid_predictions_json) -> None:
        valid_instances = [
            {"filepath": "g.jpg"},
            {"filepath": "f.jpg"},
            {"filepath": "e.jpg"},
            {"filepath": "d.jpg"},
            {"filepath": "c.jpg"},
            {"filepath": "b.jpg"},
            {"filepath": "a.jpg"},
        ]
        valid_predictions, filtered_instances = load_partial_predictions(
            valid_predictions_json, valid_instances
        )
        assert valid_predictions == {
            "e.jpg": {
                "filepath": "e.jpg",
                "country": "COUNTRY_E",
                "classifications": {
                    "classes": ["K", "L", "M"],
                    "scores": [0.9, 0.1, 0.0],
                },
                "detections": [
                    {
                        "category": "2",
                        "label": "human",
                        "conf": 0.7,
                        "bbox": [0.1, 0.2, 0.3, 0.4],
                    }
                ],
                "prediction": "K",
                "prediction_score": 0.9,
                "prediction_source": "test",
                "model_version": "0.0.0",
            },
            "f.jpg": {
                "filepath": "f.jpg",
                "classifications": {
                    "classes": ["XYZ"],
                    "scores": [0.8],
                },
                "detections": [],
                "prediction": "XYZ",
                "prediction_score": 0.4,
                "prediction_source": "test",
                "model_version": "0.0.0",
            },
        }
        assert filtered_instances == [
            {"filepath": "g.jpg"},
            {"filepath": "d.jpg"},
            {"filepath": "c.jpg"},
            {"filepath": "b.jpg"},
            {"filepath": "a.jpg"},
        ]

        with pytest.raises(RuntimeError):
            load_partial_predictions(
                valid_predictions_json,
                [{"filepath": "a.jpg"}],
            )

    @pytest.fixture
    def invalid_predictions_json(self, tmp_path) -> Generator[Path, None, None]:
        filepath = tmp_path / "invalid_predictions.json"
        with open(filepath, mode="w", encoding="utf-8") as fp:
            fp.write(
                """{
                    "predictions": [
                        {
                            "filepath": "a.jpg",
                            "countries": {
                                "USA",
                                "MEX"
                            }
                        }
                    ]
                }"""
            )
        yield filepath
        filepath.unlink()

    def test_failed_loading(self, invalid_predictions_json) -> None:
        # Python sets are not JSON serializable.
        with pytest.raises(json.JSONDecodeError):
            load_partial_predictions(invalid_predictions_json, [{"filepath": "a.jpg"}])


class TestSavePredictions:
    """Tests for the saving of predictions."""

    def test_failed_saving(self, tmp_path) -> None:
        # Python sets are not JSON serializable.
        predictions = {
            "predictions": [
                {
                    "filepath": "a.jpg",
                    "countries": {"USA", "MEX"},
                }
            ]
        }
        with pytest.raises(TypeError):
            save_predictions(predictions, tmp_path)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to run the SpeciesNet model in chunks.

This is a wrapper for run_model.py.  Supports only --classifier_only mode.
Invokes run_model using python -m speciesnet.scripts.run_model, so it assumes
that the speciesnet package is available, or that the speciesnet repo root is
on the PYTHONPATH.
"""

import json
import os
import subprocess
import sys
import tempfile
from typing import Optional
import uuid

from absl import app
from absl import flags

from speciesnet import only_one_true
from speciesnet.utils import prepare_instances_dict

# Input arguments

_INSTANCES_JSON = flags.DEFINE_string(
    "instances_json",
    None,
    "Input JSON file with instances to get predictions for.",
)
_DETECTIONS_JSON = flags.DEFINE_string(
    "detections_json",
    None,
    "Input JSON file with detections from previous runs.",
)
_FILEPATHS = flags.DEFINE_list(
    "filepaths",
    None,
    "List of image filepaths to get predictions for.",
)
_FILEPATHS_TXT = flags.DEFINE_string(
    "filepaths_txt",
    None,
    "Input TXT file with image filepaths to get predictions for.",
)
_FOLDERS = flags.DEFINE_list(
    "folders",
    None,
    "List of image folders to get predictions for.",
)
_FOLDERS_TXT = flags.DEFINE_string(
    "folders_txt",
    None,
    "Input TXT file with image folders to get predictions for.",
)

# Output arguments

_PREDICTIONS_JSON = flags.DEFINE_string(
    "predictions_json",
    None,
    "Output JSON file for storing computed predictions. Unlike run_model, if this file "
    "exists, it will be over-written.",
)

# Chunking arguments

_CHUNK_SIZE = flags.DEFINE_integer(
    "chunk_size",
    2000,
    "Number of images to run in a single process.",
)

_CLASSIFIER_ONLY = flags.DEFINE_bool(
    "classifier_only",
    False,
    "Run only the classifier component. --classifier_only enables classifier-only mode, --noclassifier_only (default) disables it.",
)
_DETECTOR_ONLY = flags.DEFINE_bool(
    "detector_only",
    False,
    "Run only the detector component. --detector_only enables detector-only mode, --nodetector_only (default) disables it.",
)
_ENSEMBLE_ONLY = flags.DEFINE_bool(
    "ensemble_only",
    False,
    "Run only the ensemble component. --ensemble_only enables ensemble-only mode, --noensemble_only (default) disables it.",
)


def _execute(
    cmd: str,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    env: Optional[dict[str, str]] = None,
    verbose: Optional[bool] = False,
):
    """Run [cmd] (a single string) in a shell, yielding each line of output to the caller."""

    os.environ["PYTHONUNBUFFERED"] = "1"

    if verbose:
        if encoding is not None:
            print(
                "Launching child process with non-default encoding {}".format(encoding)
            )
        if errors is not None:
            print(
                "Launching child process with non-default text error handling {}".format(
                    errors
                )
            )
        if env is not None:
            print(
                "Launching child process with non-default environment {}".format(
                    str(env)
                )
            )

    popen = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        universal_newlines=True,
        encoding=encoding,
        errors=errors,
        env=env,
    )
    assert popen.stdout is not None, "Process open error"
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)

    return return_code


def _execute_and_print(
    cmd: str,
    print_output: Optional[bool] = True,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    env: Optional[dict[str, str]] = None,
    verbose: Optional[bool] = False,
    catch_exceptions: Optional[bool] = True,
    echo_command: Optional[bool] = False,
):
    """Run [cmd] (a single string) in a shell, capturing and printing output.  Returns
    a dictionary with fields "status" and "output".
    """

    if echo_command:
        print("Running command:\n{}\n".format(cmd))

    to_return = {"status": -1, "output": ""}
    output = []
    try:
        for s in _execute(
            cmd, encoding=encoding, errors=errors, env=env, verbose=verbose
        ):
            output.append(s)
            if print_output:
                print(s, end="", flush=True)
        to_return["status"] = 0
    except subprocess.CalledProcessError as cpe:
        if not catch_exceptions:
            raise
        print("execute_and_print caught error: {} ({})".format(cpe.output, str(cpe)))
        to_return["status"] = cpe.returncode
    to_return["output"] = output

    return to_return


def _split_list_into_fixed_size_chunks(L, n):
    """Split the list or tuple L into chunks of size n (allowing at most one chunk with size
    less than N, i.e. len(L) does not have to be a multiple of n).

    Args:
        L (list): list to split into chunks
        n (int): preferred chunk size

    Returns:
        list: list of chunks, where each chunk is a list of length n or n-1
    """

    return [L[i * n : (i + 1) * n] for i in range((len(L) + n - 1) // n)]


def _create_argument_list_with_exceptions(all_args, exclude_args):
    """Create an argument list that excludes a subset of an original list."""

    # Filter out the excluded arguments and their values if needed
    filtered_args = []

    # Are we in the middle of an argument value list?
    skip_next = False

    for i_arg, arg in enumerate(all_args):

        # Skip this argument if flagged from previous iteration
        if skip_next:
            skip_next = False
            continue

        # Check whether this argument should be excluded
        should_exclude = False

        for exclude in exclude_args:
            # Check for exact match or key=value style argument
            if arg == exclude or arg.startswith(f"{exclude}="):
                should_exclude = True
                break

            # Check for arguments that take a value as the next argument
            if (
                (arg == exclude)
                and (i_arg < len(all_args) - 1)
                and (not all_args[i_arg + 1].startswith("-"))
            ):
                should_exclude = True
                # Skip the next value as well
                skip_next = True
                break

        if not should_exclude:
            filtered_args.append(arg)

    return filtered_args


def main(argv: str) -> None:

    if (not _CLASSIFIER_ONLY.value) or _DETECTOR_ONLY.value or _ENSEMBLE_ONLY.value:
        raise ValueError("This script only supports --classifier_only mode")

    # Check for valid inputs.
    inputs = [_INSTANCES_JSON, _FILEPATHS, _FILEPATHS_TXT, _FOLDERS, _FOLDERS_TXT]
    inputs_names = [f"--{i.name}" for i in inputs]
    inputs_values = [i.value for i in inputs]
    inputs_strings = [
        f"{name}={value}" for name, value in zip(inputs_names, inputs_values)
    ]
    if not only_one_true(*inputs_values):
        raise ValueError(
            f"Expected exactly one of [{', '.join(inputs_names)}] to be provided. "
            f"Received: [{', '.join(inputs_strings)}]."
        )

    # Extracting these out of absl flags simplifies debugging
    instances_json = _INSTANCES_JSON.value
    detections_json = _DETECTIONS_JSON.value
    filepaths = _FILEPATHS.value
    filepaths_txt = _FILEPATHS_TXT.value
    folders = _FOLDERS.value
    folders_txt = _FOLDERS_TXT.value
    chunk_size = _CHUNK_SIZE.value

    instances_dict = prepare_instances_dict(
        instances_json=instances_json,
        filepaths=filepaths,
        filepaths_txt=filepaths_txt,
        folders=folders,
        folders_txt=folders_txt,
    )

    instances = instances_dict["instances"]
    print("Loaded {} instances".format(len(instances)))

    # Split instances into chunks
    #
    # List of lists of dicts, with key "filepath"
    instance_chunks = _split_list_into_fixed_size_chunks(instances, chunk_size)
    n_chunks = len(instance_chunks)

    print("Divided instances into {} chunks".format(n_chunks))

    detections_present = False
    detection_chunks = []

    if detections_json is not None:

        detections_present = True

        with open(detections_json, "r") as f:
            detections = json.load(f)

        filepath_to_detection_results = {}
        for p in detections["predictions"]:
            filepath_to_detection_results[p["filepath"]] = p

        # Split detections into chunks that match the instance chunks
        # List of lists of dicts, with keys "filepath, detections"
        detection_chunks = []

        # chunk = instance_chunks[0]
        for chunk in instance_chunks:
            detection_chunk = []
            # instance = chunk[0]
            for instance in chunk:
                p = filepath_to_detection_results[instance["filepath"]]
                detection_chunk.append(p)
            detection_chunks.append(detection_chunk)

    else:

        print("Running model in chunks with no detections")

    # Prepare inputs for each chunk
    chunk_file_base = os.path.join(
        tempfile.gettempdir(), "speciesnet_chunks", str(uuid.uuid1())
    )
    os.makedirs(chunk_file_base, exist_ok=True)

    detection_chunk_files = []
    instances_chunk_files = []

    # Double-check that we lined up the instances and detections correctly, and
    # prepare outout files
    for i_chunk, instance_chunk in enumerate(instance_chunks):

        instance_paths = set(instance["filepath"] for instance in instance_chunk)

        if detections_present:

            detection_chunk = detection_chunks[i_chunk]
            assert instance_paths == set(
                detection["filepath"] for detection in detection_chunk
            )
            detection_chunk_file = os.path.join(
                chunk_file_base,
                "detections_chunk_{}.json".format(str(i_chunk).zfill(4)),
            )
            detection_chunk_dict = {"predictions": detection_chunk}
            with open(detection_chunk_file, "w") as f:
                json.dump(detection_chunk_dict, f, indent=1)
            detection_chunk_files.append(detection_chunk_file)

        instance_chunk_file = os.path.join(
            chunk_file_base, "instances_chunk_{}.json".format(str(i_chunk).zfill(4))
        )

        instances_chunk_dict = {"instances": instance_chunk}

        with open(instance_chunk_file, "w") as f:
            json.dump(instances_chunk_dict, f, indent=1)

        instances_chunk_files.append(instance_chunk_file)

    # Assemble commands for new processes
    classification_chunk_files = []
    classification_commands = []

    # These are the arguments we don't want to pass to child processes
    exclude_args = [
        "--predictions_json",
        "--instances_json",
        "--detections_json",
        "--filepaths",
        "--filepaths_txt",
        "--folders",
        "--chunk_size",
    ]
    all_args = sys.argv

    filtered_args = _create_argument_list_with_exceptions(all_args, exclude_args)

    for i_chunk in range(0, n_chunks):

        chunk_cmd = "python -m speciesnet.scripts.run_model"
        for arg in filtered_args:
            chunk_cmd += f" {arg}"
        instance_chunk_file = instances_chunk_files[i_chunk]
        classification_chunk_file = os.path.join(
            chunk_file_base,
            "classifications_chunk_{}.json".format(str(i_chunk).zfill(4)),
        )
        classification_chunk_files.append(classification_chunk_file)
        chunk_cmd += f" --instances_json {instance_chunk_file}"
        if detections_present:
            detection_chunk_file = detection_chunk_files[i_chunk]
            chunk_cmd += f" --detections_json {detection_chunk_file}"
        else:
            chunk_cmd += " --bypass_prompts"
        chunk_cmd += f" --predictions_json {classification_chunk_file}"
        classification_commands.append(chunk_cmd)

    # Run commands
    for i_chunk in range(0, n_chunks):
        chunk_cmd = classification_commands[i_chunk]
        print("Running command for chunk {}: {}".format(i_chunk, chunk_cmd))
        _execute_and_print(classification_commands[i_chunk])
        assert os.path.isfile(classification_chunk_files[i_chunk])

    # Merge results
    output_predictions = []
    for i_chunk in range(0, n_chunks):
        with open(classification_chunk_files[i_chunk], "r") as f:
            chunk_results = json.load(f)
        output_predictions.extend(chunk_results["predictions"])

    print("Read {} predictions".format(len(output_predictions)))

    output_dict = {"predictions": output_predictions}
    with open(_PREDICTIONS_JSON.value, "w") as f:
        json.dump(output_dict, f, indent=1)


if __name__ == "__main__":
    app.run(main, flags_parser=lambda args: flags.FLAGS(args, known_only=True))



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-module-docstring

from speciesnet.geolocation import find_admin1_region as admin1


class TestFindAdmin1Region:
    """Tests for the first-level administrative division."""

    def test_missing_country(self) -> None:
        assert admin1(country=None, latitude=None, longitude=None) is None
        assert admin1(country=None, latitude=0, longitude=0) is None
        assert admin1(country=None, admin1_region="AK") is None

    def test_given_admin1_region(self) -> None:
        assert admin1(country="USA", admin1_region="HI") == "HI"
        assert (
            admin1(country="USA", admin1_region="CA", latitude=37, longitude=-119)
            == "CA"
        )
        assert (
            admin1(country="USA", admin1_region="CA", latitude=46.01, longitude=-69.19)
            == "CA"
        )
        assert (
            admin1(country="USA", admin1_region="DC", latitude=46.01, longitude=-69.19)
            == "DC"
        )

    def test_usa_states(self) -> None:
        assert admin1(country="USA", latitude=None, longitude=None) is None
        assert admin1(country="USA", latitude=0, longitude=None) is None
        assert admin1(country="USA", latitude=None, longitude=0) is None
        assert admin1(country="USA", latitude=0, longitude=0) is None
        assert admin1(country="USA", latitude=-23, longitude=133) is None
        assert admin1(country="USA", latitude=45, longitude=0) is None

        assert admin1(country="USA", latitude=65, longitude=-150) == "AK"
        assert admin1(country="USA", latitude=19, longitude=-155) == "HI"
        assert admin1(country="USA", latitude=37, longitude=-119) == "CA"
        assert admin1(country="USA", latitude=46.01, longitude=-69.19) == "ME"
        assert admin1(country="USA", latitude=38.63, longitude=-103.43) == "CO"
        assert admin1(country="USA", latitude=38.92, longitude=-77.01) == "DC"

    def test_usa_territories(self) -> None:
        assert admin1(country="ASM", latitude=-14.23, longitude=-169.46) is None
        assert admin1(country="USA", latitude=-14.23, longitude=-169.46) == "AS"
        assert admin1(country="GUM", latitude=13.42, longitude=144.71) is None
        assert admin1(country="USA", latitude=13.42, longitude=144.71) == "GU"
        assert admin1(country="MNP", latitude=16.34, longitude=145.66) is None
        assert admin1(country="USA", latitude=16.34, longitude=145.66) == "MP"
        assert admin1(country="PRI", latitude=18.22, longitude=-66.67) is None
        assert admin1(country="USA", latitude=18.22, longitude=-66.67) == "PR"
        assert admin1(country="UMI", latitude=19.28, longitude=166.64) is None
        # "UM" is not currently supported by the reverse geocoder.
        # assert admin1(country="USA", latitude=19.28, longitude=166.64) == "UM"
        assert admin1(country="USA", latitude=19.28, longitude=166.64) is None
        assert admin1(country="VIR", latitude=18.35, longitude=-64.71) is None
        assert admin1(country="USA", latitude=18.35, longitude=-64.71) == "VI"

    def test_other_countries(self) -> None:
        assert admin1(country="AUS", latitude=None, longitude=None) is None
        assert admin1(country="AUS", latitude=-23, longitude=133) is None
        assert admin1(country="BRA", latitude=0, longitude=0) is None
        assert admin1(country="BRA", latitude=-8, longitude=-56) is None



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Taxonomy-related utility functions.

Provides functions for working with taxonomic labels, such as finding ancestors at
specific levels and extracting full class strings.
"""

from typing import Optional

from speciesnet.constants import Classification


def get_ancestor_at_level(
    label: str, taxonomy_level: str, taxonomy_map: dict
) -> Optional[str]:
    """Finds the taxonomy item corresponding to a label's ancestor at a given level.

    E.g. The ancestor at family level for
    `uuid;class;order;family;genus;species;common_name` is
    `another_uuid;class;order;family;;;another_common_name`.

    Args:
        label:
            String label for which to find the ancestor.
        taxonomy_level:
            One of "species", "genus", "family", "order", "class" or "kingdom",
            indicating the taxonomy level at which to find a label's ancestor.
        taxonomy_map:
            Dictionary mapping taxa to labels.

    Returns:
        A string label indicating the ancestor at the requested taxonomy level. In
        case the taxonomy doesn't contain the corresponding ancestor, return `None`.

    Raises:
        ValueError:
            If the given label is invalid.
    """

    label_parts = label.split(";")
    if len(label_parts) != 7:
        raise ValueError(
            f"Expected label made of 7 parts, but found only {len(label_parts)}: "
            f"{label}"
        )

    if taxonomy_level == "species":
        ancestor_parts = label_parts[1:6]
        if not ancestor_parts[4]:
            return None
    elif taxonomy_level == "genus":
        ancestor_parts = label_parts[1:5] + [""]
        if not ancestor_parts[3]:
            return None
    elif taxonomy_level == "family":
        ancestor_parts = label_parts[1:4] + ["", ""]
        if not ancestor_parts[2]:
            return None
    elif taxonomy_level == "order":
        ancestor_parts = label_parts[1:3] + ["", "", ""]
        if not ancestor_parts[1]:
            return None
    elif taxonomy_level == "class":
        ancestor_parts = label_parts[1:2] + ["", "", "", ""]
        if not ancestor_parts[0]:
            return None
    elif taxonomy_level == "kingdom":
        ancestor_parts = ["", "", "", "", ""]
        if not label_parts[1] and label != Classification.ANIMAL:
            return None
    else:
        return None

    ancestor = ";".join(ancestor_parts)
    return taxonomy_map.get(ancestor)


def get_full_class_string(label: str) -> str:
    """Extracts the full class string corresponding to a given label.

    E.g. The full class string for the label
    `uuid;class;order;family;genus;species;common_name` is
    `class;order;family;genus;species`.

    Args:
        label:
            String label for which to extract the full class string.

    Returns:
        Full class string for the given label.

    Raises:
        ValueError: If the given label is invalid.
    """

    label_parts = label.split(";")
    if len(label_parts) != 7:
        raise ValueError(
            f"Expected label made of 7 parts, but found only {len(label_parts)}: "
            f"{label}"
        )
    return ";".join(label_parts[1:6])



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Geolocation utilities."""

__all__ = [
    "find_admin1_region",
]

from typing import Optional

import reverse_geocoder as rg

US_COUNTRY_CODES_ALPHA_2_TO_ALPHA_3 = {  # ISO 3166-1 Alpha-2 => ISO 3166-1 Alpha-3
    "US": "USA",  # United States of America
    "AS": "ASM",  # American Samoa
    "GU": "GUM",  # Guam
    "MP": "MNP",  # Northern Mariana Islands
    "PR": "PRI",  # Puerto Rico
    "UM": "UMI",  # United States Minor Outlying Islands
    "VI": "VIR",  # United States Virgin Islands
}
US_COUNTRY_CODES_ALPHA_2 = US_COUNTRY_CODES_ALPHA_2_TO_ALPHA_3.keys()
US_COUNTRY_CODES_ALPHA_3 = US_COUNTRY_CODES_ALPHA_2_TO_ALPHA_3.values()

US_ADMIN1_REGION_TO_ABBREVIATION = {
    "Alabama": "AL",
    "Alaska": "AK",
    "Arizona": "AZ",
    "Arkansas": "AR",
    "California": "CA",
    "Colorado": "CO",
    "Connecticut": "CT",
    "Delaware": "DE",
    "Florida": "FL",
    "Georgia": "GA",
    "Hawaii": "HI",
    "Idaho": "ID",
    "Illinois": "IL",
    "Indiana": "IN",
    "Iowa": "IA",
    "Kansas": "KS",
    "Kentucky": "KY",
    "Louisiana": "LA",
    "Maine": "ME",
    "Maryland": "MD",
    "Massachusetts": "MA",
    "Michigan": "MI",
    "Minnesota": "MN",
    "Mississippi": "MS",
    "Missouri": "MO",
    "Montana": "MT",
    "Nebraska": "NE",
    "Nevada": "NV",
    "New Hampshire": "NH",
    "New Jersey": "NJ",
    "New Mexico": "NM",
    "New York": "NY",
    "North Carolina": "NC",
    "North Dakota": "ND",
    "Ohio": "OH",
    "Oklahoma": "OK",
    "Oregon": "OR",
    "Pennsylvania": "PA",
    "Rhode Island": "RI",
    "South Carolina": "SC",
    "South Dakota": "SD",
    "Tennessee": "TN",
    "Texas": "TX",
    "Utah": "UT",
    "Vermont": "VT",
    "Virginia": "VA",
    "Washington": "WA",
    "West Virginia": "WV",
    "Wisconsin": "WI",
    "Wyoming": "WY",
    "District of Columbia": "DC",
    # "District of Columbia" appears with an alternative name in the reverse geocoder.
    "Washington, D.C.": "DC",
}


ADMIN1_REGION_SUPPORTED_COUNTRIES = US_COUNTRY_CODES_ALPHA_3


def find_admin1_region(
    country: Optional[str] = None,
    admin1_region: Optional[str] = None,
    latitude: Optional[float] = None,
    longitude: Optional[float] = None,
) -> Optional[str]:
    """Finds the first-level administrative division for a given (lat, lon) location.

    Skips doing any work if both `country` and `admin1_region` are already specified.

    Currently implemented only for USA and its territories:
        * If the country is "USA" and the (lat, lon) coordinates fall within a USA state
          or USA territory, the admin1 region will be the two-letter country code of
          that state or territory (e.g. "NY" or "PR").
        * If the country is a USA territory country code (e.g. "PRI") and the (lat, lon)
          coordinates fall within that territory, the admin1 region will be `None`.

    Args:
        country:
            Country in ISO 3166-1 alpha-3 format. Optional.
        admin1_region:
            First-level administrative division in ISO 3166-2 format. Optional.
        latitude:
            Float value representing latitude. Optional.
        longitude:
            Float value representing longitude. Optional.

    Returns:
        First-level administrative division in ISO 3166-2 format. If no matching admin1
        region is found (e.g. due to missing country, latitude or longitude, or due to
        country not supported), return `None`.
    """

    if country and admin1_region:
        return admin1_region

    if (
        country not in ADMIN1_REGION_SUPPORTED_COUNTRIES
        or latitude is None
        or longitude is None
    ):
        return None

    if country in US_COUNTRY_CODES_ALPHA_3:
        if country != "USA":
            return None

        result = rg.search((latitude, longitude), mode=1, verbose=False)[0]
        if result["cc"] not in US_COUNTRY_CODES_ALPHA_2:
            return None
        if result["cc"] == "US":
            if (
                "admin1" in result
                and result["admin1"] in US_ADMIN1_REGION_TO_ABBREVIATION
            ):
                return US_ADMIN1_REGION_TO_ABBREVIATION[result["admin1"]]
            else:
                return None
        else:
            return result["cc"]

    return None



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ensemble functionality of SpeciesNet."""

__all__ = [
    "SpeciesNetEnsemble",
]

import json
import time
from typing import Any, Callable

from absl import logging
from humanfriendly import format_timespan

from speciesnet.constants import Classification
from speciesnet.constants import Failure
from speciesnet.ensemble_prediction_combiner import combine_predictions_for_single_item
from speciesnet.geofence_utils import geofence_animal_classification
from speciesnet.geofence_utils import roll_up_labels_to_first_matching_level
from speciesnet.utils import ModelInfo

# Handy type aliases.
PredictionLabelType = str
PredictionScoreType = float
PredictionSourceType = str
PredictionType = tuple[PredictionLabelType, PredictionScoreType, PredictionSourceType]


class SpeciesNetEnsemble:
    """Ensemble component of SpeciesNet."""

    def __init__(
        self,
        model_name: str,
        geofence: bool = True,
        prediction_combiner: Callable = combine_predictions_for_single_item,
    ) -> None:
        """Loads the ensemble resources.

        Args:
            model_name:
                String value identifying the model to be loaded. It can be a Kaggle
                identifier (starting with `kaggle:`), a HuggingFace identifier (starting
                with `hf:`) or a local folder to load the model from.
            geofence:
                Whether to enable geofencing. If `False` skip it entirely.
        """

        start_time = time.time()

        self.model_info = ModelInfo(model_name)
        self.enable_geofence = geofence
        self.taxonomy_map = self.load_taxonomy()
        self.geofence_map = self.load_geofence()
        self.prediction_combiner = prediction_combiner

        end_time = time.time()
        logging.info(
            "Loaded SpeciesNetEnsemble in %s.",
            format_timespan(end_time - start_time),
        )

    def load_taxonomy(self):
        """Loads the taxonomy from the model info."""

        def _taxa_from_label(label: str) -> str:
            return ";".join(label.split(";")[1:6])

        # Create taxonomy map.
        with open(self.model_info.taxonomy, mode="r", encoding="utf-8") as fp:
            labels = [line.strip() for line in fp.readlines()]
            taxonomy_map = {_taxa_from_label(label): label for label in labels}

            for label in [
                Classification.BLANK,
                Classification.VEHICLE,
                Classification.UNKNOWN,
            ]:
                taxa = _taxa_from_label(label)
                if taxa in taxonomy_map:
                    del taxonomy_map[taxa]

            for label in [Classification.HUMAN, Classification.ANIMAL]:
                taxa = _taxa_from_label(label)
                taxonomy_map[taxa] = label
        return taxonomy_map

    def load_geofence(self):
        """Loads the geofence map from the model info."""

        with open(self.model_info.geofence, mode="r", encoding="utf-8") as fp:
            geofence_map = json.load(fp)
        return geofence_map

    def combine(  # pylint: disable=too-many-positional-arguments
        self,
        filepaths: list[str],
        classifier_results: dict[str, Any],
        detector_results: dict[str, Any],
        geolocation_results: dict[str, Any],
        partial_predictions: dict[str, dict],
    ) -> list[dict[str, Any]]:
        """Ensembles classifications and detections for a list of images.

        Args:
            filepaths:
                List of filepaths to ensemble predictions for.
            classifier_results:
                Dict of classifier results, with keys given by the filepaths to ensemble
                predictions for.
            detector_results:
                Dict of detector results, with keys given by the filepaths to ensemble
                predictions for.
            geolocation_results:
                Dict of geolocation results, with keys given by the filepaths to
                ensemble predictions for.
            partial_predictions:
                Dict of partial predictions from previous ensemblings, with keys given
                by the filepaths for which predictions where already ensembled. Used to
                skip re-ensembling for the matching filepaths.

        Returns:
            List of ensembled predictions.
        """

        results = []
        for filepath in filepaths:
            # Use the result from previously computed predictions when available.
            if filepath in partial_predictions:
                results.append(partial_predictions[filepath])
                continue

            # Check for failures.
            failure = Failure(0)
            if (
                filepath in classifier_results
                and "failures" not in classifier_results[filepath]
            ):
                classifications = classifier_results[filepath]["classifications"]
            else:
                classifications = None
                failure |= Failure.CLASSIFIER
            if (
                filepath in detector_results
                and "failures" not in detector_results[filepath]
            ):
                detections = detector_results[filepath]["detections"]
            else:
                detections = None
                failure |= Failure.DETECTOR
            if filepath in geolocation_results:
                geolocation = geolocation_results[filepath]
            else:
                geolocation = {}
                failure |= Failure.GEOLOCATION

            # Add as much raw information as possible to the prediction result.
            result = {
                "filepath": filepath,
                "failures": (
                    [f.name for f in Failure if f in failure] if failure else None
                ),
                "country": geolocation.get("country"),
                "admin1_region": geolocation.get("admin1_region"),
                "latitude": geolocation.get("latitude"),
                "longitude": geolocation.get("longitude"),
                "classifications": classifications,
                "detections": detections,
            }
            result = {key: value for key, value in result.items() if value is not None}

            # Most importantly, ensemble everything into a single prediction.
            if classifications is not None and detections is not None:
                prediction, score, source = self.prediction_combiner(
                    classifications=classifications,
                    detections=detections,
                    country=geolocation.get("country"),
                    admin1_region=geolocation.get("admin1_region"),
                    taxonomy_map=self.taxonomy_map,
                    geofence_map=self.geofence_map,
                    enable_geofence=self.enable_geofence,
                    geofence_fn=geofence_animal_classification,
                    roll_up_fn=roll_up_labels_to_first_matching_level,
                )
                result["prediction"] = (
                    prediction.value
                    if isinstance(prediction, Classification)
                    else prediction
                )
                result["prediction_score"] = score
                result["prediction_source"] = source

            # Finally, report the model version.
            result["model_version"] = self.model_info.version

            results.append(result)

        return results



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-module-docstring
# pylint: disable=protected-access
# pylint: disable=redefined-outer-name

import pytest

from speciesnet.constants import Classification
from speciesnet.constants import Detection
from speciesnet.ensemble_prediction_combiner import combine_predictions_for_single_item

# fmt: off
# pylint: disable=line-too-long

BLANK = "f1856211-cfb7-4a5b-9158-c0f72fd09ee6;;;;;;blank"
BLANK_FC = ";;;;"
HUMAN = "990ae9dd-7a59-4344-afcb-1b7b21368000;mammalia;primates;hominidae;homo;sapiens;human"
HUMAN_FC = "mammalia;primates;hominidae;homo;sapiens"
VEHICLE = "e2895ed5-780b-48f6-8a11-9e27cb594511;;;;;;vehicle"
VEHICLE_FC = ";;;;"

LION = "ddf59264-185a-4d35-b647-2785792bdf54;mammalia;carnivora;felidae;panthera;leo;lion"
LION_FC = "mammalia;carnivora;felidae;panthera;leo"
PANTHERA_GENUS = "fbb23d07-6677-43db-b650-f99ac452c50f;mammalia;carnivora;felidae;panthera;;panthera species"
PANTHERA_GENUS_FC = "mammalia;carnivora;felidae;panthera;"
FELIDAE_FAMILY = "df8514b0-10a5-411f-8ed6-0f415e8153a3;mammalia;carnivora;felidae;;;cat family"
FELIDAE_FAMILY_FC = "mammalia;carnivora;felidae;;"
CARNIVORA_ORDER = "eeeb5d26-2a47-4d01-a3de-10b33ec0aee4;mammalia;carnivora;;;;carnivorous mammal"
CARNIVORA_ORDER_FC = "mammalia;carnivora;;;"
MAMMALIA_CLASS = "f2d233e3-80e3-433d-9687-e29ecc7a467a;mammalia;;;;;mammal"
MAMMALIA_CLASS_FC = "mammalia;;;;"
ANIMAL_KINGDOM = "1f689929-883d-4dae-958c-3d57ab5b6c16;;;;;;animal"
ANIMAL_KINGDOM_FC = ";;;;"

BROWN_BEAR = "330bb1e9-84d6-4e41-afa9-938aee17ea29;mammalia;carnivora;ursidae;ursus;arctos;brown bear"
BROWN_BEAR_FC = "mammalia;carnivora;ursidae;ursus;arctos"
POLAR_BEAR = "e7f83bf6-df2c-4ce0-97fc-2f233df23ec4;mammalia;carnivora;ursidae;ursus;maritimus;polar bear"
POLAR_BEAR_FC = "mammalia;carnivora;ursidae;ursus;maritimus"
GIANT_PANDA = "85662682-67c1-4ecb-ba05-ba12e2df6b65;mammalia;carnivora;ursidae;ailuropoda;melanoleuca;giant panda"
GIANT_PANDA_FC = "mammalia;carnivora;ursidae;ailuropoda;melanoleuca"
URSUS_GENUS = "5a0f5e3f-c634-4b86-910a-b105cb526a24;mammalia;carnivora;ursidae;ursus;;ursus species"
URSUS_GENUS_FC = "mammalia;carnivora;ursidae;ursus;"
URSIDAE_FAMILY = "ec1a70f4-41c0-4aba-9150-292fb2b7a324;mammalia;carnivora;ursidae;;;bear family"
URSIDAE_FAMILY_FC = "mammalia;carnivora;ursidae;;"

PUMA = "9c564562-9429-405c-8529-04cff7752282;mammalia;carnivora;felidae;puma;concolor;puma"
PUMA_FC = "mammalia;carnivora;felidae;puma;concolor"
SAND_CAT = "e588253d-d61d-4149-a96c-8c245927a80f;mammalia;carnivora;felidae;felis;margarita;sand cat"
SAND_CAT_FC = "mammalia;carnivora;felidae;felis;margarita"

# pylint: enable=line-too-long
# fmt: on


@pytest.fixture
def taxonomy_map():
    return {
        BLANK_FC: BLANK,
        HUMAN_FC: HUMAN,
        VEHICLE_FC: VEHICLE,
        LION_FC: LION,
        PANTHERA_GENUS_FC: PANTHERA_GENUS,
        FELIDAE_FAMILY_FC: FELIDAE_FAMILY,
        CARNIVORA_ORDER_FC: CARNIVORA_ORDER,
        MAMMALIA_CLASS_FC: MAMMALIA_CLASS,
        ANIMAL_KINGDOM_FC: ANIMAL_KINGDOM,
        BROWN_BEAR_FC: BROWN_BEAR,
        POLAR_BEAR_FC: POLAR_BEAR,
        GIANT_PANDA_FC: GIANT_PANDA,
        URSUS_GENUS_FC: URSUS_GENUS,
        URSIDAE_FAMILY_FC: URSIDAE_FAMILY,
    }


@pytest.fixture
def geofence_map():
    return {
        LION_FC: {
            "allow": {
                "KEN": [],
                "TZA": [],
            }
        },
        PANTHERA_GENUS_FC: {
            "allow": {
                "KEN": [],
                "TZA": [],
                "USA": ["AK", "CA"],
            }
        },
        FELIDAE_FAMILY_FC: {
            "allow": {
                "FRA": [],
                "KEN": [],
                "TZA": [],
                "USA": [],
            },
            "block": {
                "FRA": [],
                "USA": ["NY"],
            },
        },
        SAND_CAT_FC: {
            "block": {
                "AUS": [],
            },
        },
        URSIDAE_FAMILY_FC: {
            "block": {
                "GBR": [],
            },
        },
    }


def mock_geofence_fn(*args, **kwargs):
    del args
    del kwargs
    return "geofenced_label", 0.9, "geofence_source"


def mock_roll_up_fn(*args, **kwargs):
    del args
    del kwargs
    return "rollup_label", 0.8, "rollup_source"


def mock_roll_up_fn_return_none(*args, **kwargs):
    del args
    del kwargs


class TestPredictionEnsembleCombiner:
    """Tests for the ensemble combiner function."""

    def test_combine_predictions_for_single_item_human_detections(
        self, taxonomy_map, geofence_map
    ) -> None:
        # High-confidence HUMAN detection.
        classifications = {"classes": [LION], "scores": [0.1]}
        detections = [{"label": Detection.HUMAN, "conf": 0.8}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == (Classification.HUMAN, 0.8, "detector")

        # Mid-confidence HUMAN detection + high-confidence HUMAN classification.
        classifications = {"classes": [HUMAN], "scores": [0.7]}
        detections = [{"label": Detection.HUMAN, "conf": 0.3}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == (Classification.HUMAN, 0.7, "classifier")

        # Mid-confidence HUMAN detection + high-confidence VEHICLE classification.
        classifications = {"classes": [VEHICLE], "scores": [0.7]}
        detections = [{"label": Detection.HUMAN, "conf": 0.3}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == (Classification.HUMAN, 0.7, "classifier")

    def test_combine_predictions_for_single_item_vehicle_detections(
        self, taxonomy_map, geofence_map
    ) -> None:
        # Mid-confidence VEHICLE detection + high-confidence HUMAN classification.
        classifications = {"classes": [HUMAN], "scores": [0.7]}
        detections = [{"label": Detection.VEHICLE, "conf": 0.3}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == (Classification.HUMAN, 0.7, "classifier")

        # High-confidence VEHICLE detection.
        classifications = {"classes": [LION], "scores": [0.1]}
        detections = [{"label": Detection.VEHICLE, "conf": 0.8}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == (Classification.VEHICLE, 0.8, "detector")

        # Mid-confidence VEHICLE detection + high-confidence VEHICLE classification.
        classifications = {"classes": [VEHICLE], "scores": [0.5]}
        detections = [{"label": Detection.VEHICLE, "conf": 0.3}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == (Classification.VEHICLE, 0.5, "classifier")

    def test_combine_predictions_for_single_item_blank_classifications(
        self, taxonomy_map, geofence_map
    ) -> None:
        # High-confidence BLANK "detection" + high-confidence BLANK classification.
        classifications = {"classes": [BLANK], "scores": [0.7]}
        detections = [{"label": Detection.ANIMAL, "conf": 0.1}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == (Classification.BLANK, 0.7, "classifier")

        # Extra-high-confidence BLANK classification.
        classifications = {"classes": [BLANK], "scores": [0.995]}
        detections = [{"label": Detection.ANIMAL, "conf": 0.3}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == (Classification.BLANK, 0.995, "classifier")

    def test_combine_predictions_for_single_item_animal_classifications(
        self, taxonomy_map, geofence_map
    ) -> None:
        # Extra-high-confidence ANIMAL classification.
        classifications = {"classes": [LION], "scores": [0.9]}
        detections = [{"label": Detection.ANIMAL, "conf": 0.1}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == ("geofenced_label", 0.9, "geofence_source")

        # High-confidence ANIMAL classification + mid-confidence ANIMAL detection.
        classifications = {"classes": [LION], "scores": [0.7]}
        detections = [{"label": Detection.ANIMAL, "conf": 0.3}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == ("geofenced_label", 0.9, "geofence_source")

    def test_combine_predictions_for_single_item_animal_rollups(
        self, taxonomy_map, geofence_map
    ) -> None:
        # High-confidence ANIMAL rollups.
        classifications = {"classes": [LION], "scores": [0.6]}
        detections = [{"label": Detection.ANIMAL, "conf": 0.1}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn,
        )
        assert result == ("rollup_label", 0.8, "rollup_source")

    def test_combine_predictions_for_single_item_animal_detections(
        self, taxonomy_map, geofence_map
    ) -> None:
        # Mid-confidence ANIMAL detection.
        classifications = {"classes": [LION], "scores": [0.1]}
        detections = [{"label": Detection.ANIMAL, "conf": 0.6}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn_return_none,
        )
        assert result == (Classification.ANIMAL, 0.6, "detector")

    def test_combine_predictions_for_single_item_unknown(
        self, taxonomy_map, geofence_map
    ) -> None:
        # Fallback to UNKNOWN classification.
        classifications = {"classes": [LION], "scores": [0.1]}
        detections = [{"label": Detection.ANIMAL, "conf": 0.1}]
        result = combine_predictions_for_single_item(
            classifications=classifications,
            detections=detections,
            country=None,
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
            geofence_fn=mock_geofence_fn,
            roll_up_fn=mock_roll_up_fn_return_none,
        )
        assert result == (Classification.UNKNOWN, 0.1, "classifier")



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Geofence-related utility functions.

Provides functions for checking geofencing rules and rolling up
classification labels based on geographic restrictions and taxonomic levels.
"""

from typing import Optional

from speciesnet.constants import Classification
from speciesnet.taxonomy_utils import get_ancestor_at_level
from speciesnet.taxonomy_utils import get_full_class_string

# Handy type aliases.
PredictionLabelType = str
PredictionScoreType = float
PredictionSourceType = str
PredictionType = tuple[PredictionLabelType, PredictionScoreType, PredictionSourceType]


def should_geofence_animal_classification(
    label: str,
    country: Optional[str],
    admin1_region: Optional[str],
    geofence_map: dict,
    enable_geofence: bool,
) -> bool:
    """Checks whether to geofence animal prediction in a country or admin1_region.

    Args:
        label:
            Animal label to check geofence rules for.
        country:
            Country (in ISO 3166-1 alpha-3 format) to check geofence rules for.
            Optional.
        admin1_region:
            First-level administrative division (in ISO 3166-2 format) to check
            geofence rules for. Optional.
        geofence_map:
            Dictionary mapping full class strings to geofence rules.
        enable_geofence:
            Whether geofencing is enabled.

    Returns:
        A boolean indicating whether to geofence given animal prediction.
    """

    # Do not geofence if geofencing is disabled.
    if not enable_geofence:
        return False

    # Do not geofence if country was not provided.
    if not country:
        return False

    # Do not geofence if full class string is missing from the geofence map.
    full_class_string = get_full_class_string(label)
    if full_class_string not in geofence_map:
        return False

    # Check if we need to geofence based on "allow" rules.
    allow_countries = geofence_map[full_class_string].get("allow")
    if allow_countries:
        if country not in allow_countries:
            # Geofence when country was not explicitly allowed.
            return True
        else:
            allow_admin1_regions = allow_countries[country]
            if (
                admin1_region
                and allow_admin1_regions
                and admin1_region not in allow_admin1_regions
            ):
                # Geofence when admin1_region was not explicitly allowed.
                return True

    # Check if we need to geofence based on "block" rules.
    block_countries = geofence_map[full_class_string].get("block")
    if block_countries:
        if country in block_countries:
            block_admin1_regions = block_countries[country]
            if not block_admin1_regions:
                # Geofence when entire country was blocked.
                return True
            elif admin1_region and admin1_region in block_admin1_regions:
                # Geofence when admin1_region was blocked.
                return True

    # Do not geofence if no rule enforced that.
    return False


def roll_up_labels_to_first_matching_level(  # pylint: disable=too-many-positional-arguments
    labels: list[str],
    scores: list[float],
    country: Optional[str],
    admin1_region: Optional[str],
    target_taxonomy_levels: list[str],
    non_blank_threshold: float,
    taxonomy_map: dict,
    geofence_map: dict,
    enable_geofence: bool,
) -> Optional[PredictionType]:
    """Rolls up prediction labels to the first taxonomy level above given threshold.

    Args:
        labels:
            List of classification labels.
        scores:
            List of classification scores.
        country:
            Country (in ISO 3166-1 alpha-3 format) associated with prediction.
            Optional.
        admin1_region:
            First-level administrative division (in ISO 3166-2 format) associated
            with prediction. Optional.
        target_taxonomy_levels:
            Ordered list of taxonomy levels at which to roll up classification
            labels and check if the cumulative score passes the given threshold.
            Levels must be a subset of: "species", "genus", "family", "order",
            "class", "kingdom".
        non_blank_threshold:
            Min threshold at which the cumulative score is good enough to consider
            the rollup successful.
        taxonomy_map:
            Dictionary mapping taxa to labels.
        geofence_map:
            Dictionary mapping full class strings to geofence rules.
        enable_geofence:
            Whether geofencing is enabled.

    Returns:
        A tuple of <label, score, prediction_source> describing the first taxonomy
        level at which the cumulative score passes the given threshold. If no such
        level exists, return `None`.

    Raises:
        ValueError:
            If the taxonomy level if not one of: "species", "genus", "family",
            "order", "class", "kingdom".
    """

    expected_target_taxonomy_levels = {
        "species",
        "genus",
        "family",
        "order",
        "class",
        "kingdom",
    }
    unknown_target_taxonomy_levels = set(target_taxonomy_levels).difference(
        expected_target_taxonomy_levels
    )
    if unknown_target_taxonomy_levels:
        raise ValueError(
            "Unexpected target taxonomy level(s): "
            f"{unknown_target_taxonomy_levels}. "
            f"Expected only levels from the set: {expected_target_taxonomy_levels}."
        )

    # Accumulate scores at each taxonomy level and, if they pass the desired
    # threshold, return that rollup label.
    for taxonomy_level in target_taxonomy_levels:
        accumulated_scores = {}
        for label, score in zip(labels, scores):
            rollup_label = get_ancestor_at_level(
                label=label, taxonomy_level=taxonomy_level, taxonomy_map=taxonomy_map
            )
            if rollup_label:
                new_score = accumulated_scores.get(rollup_label, 0.0) + score
                accumulated_scores[rollup_label] = new_score

        max_rollup_label = None
        max_rollup_score = 0.0
        for rollup_label, rollup_score in accumulated_scores.items():
            if (
                rollup_score > max_rollup_score
                and not should_geofence_animal_classification(
                    rollup_label, country, admin1_region, geofence_map, enable_geofence
                )
            ):
                max_rollup_label = rollup_label
                max_rollup_score = rollup_score
        if max_rollup_score > non_blank_threshold and max_rollup_label:
            return (
                max_rollup_label,
                max_rollup_score,
                f"classifier+rollup_to_{taxonomy_level}",
            )

    return None


def geofence_animal_classification(
    *,
    labels: list[str],
    scores: list[float],
    country: Optional[str],
    admin1_region: Optional[str],
    taxonomy_map: dict,
    geofence_map: dict,
    enable_geofence: bool,
) -> PredictionType:
    """Geofences animal prediction in a country or admin1_region.

    Under the hood, this also rolls up the labels every time it encounters a
    geofenced label.

    Args:
        labels:
            List of classification labels.
        scores:
            List of classification scores.
        country:
            Country (in ISO 3166-1 alpha-3 format) associated with prediction.
            Optional.
        admin1_region:
            First-level administrative division (in ISO 3166-2 format) associated
            with prediction. Optional.
        taxonomy_map:
            Dictionary mapping taxa to labels.
        geofence_map:
            Dictionary mapping full class strings to geofence rules.
        enable_geofence:
            Whether geofencing is enabled.

    Returns:
        A tuple of <label, score, prediction_source> describing the result of the
        combined geofence and rollup operations.
    """

    if should_geofence_animal_classification(
        labels[0], country, admin1_region, geofence_map, enable_geofence
    ):
        rollup = roll_up_labels_to_first_matching_level(
            labels=labels,
            scores=scores,
            country=country,
            admin1_region=admin1_region,
            target_taxonomy_levels=["family", "order", "class", "kingdom"],
            # Force the rollup to pass the top classification score.
            non_blank_threshold=scores[0] - 1e-10,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=enable_geofence,
        )
        if rollup:
            rollup_label, rollup_score, rollup_source = rollup
            return (
                rollup_label,
                rollup_score,
                "classifier+geofence+" + rollup_source[len("classifier+") :],
            )
        else:
            # Normally, this return statement should never be reached since the
            # animal rollup would eventually succeed (even though that may be at
            # "kingdom" level, as a last resort). The only scenario when this could
            # still be reached is if the method was incorrectly called with a list
            # of non-animal labels (e.g. blanks, vehicles). In this case it's best
            # to return an unknown classification, while propagating the top score.
            return (
                Classification.UNKNOWN,
                scores[0],
                "classifier+geofence+rollup_failed",
            )
    else:
        return labels[0], scores[0], "classifier"



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-module-docstring

import json
import logging
from typing import Any, Optional

import pytest

from speciesnet.multiprocessing import SpeciesNet


def assert_approx_objs(
    obj1: Any,
    obj2: Any,
    rtol: Optional[float] = None,  # Relative tolerance.
    atol: Optional[float] = None,  # Absolute tolerance.
) -> None:
    if isinstance(obj1, dict) and isinstance(obj2, dict):
        assert set(obj1.keys()) == set(obj2.keys())
        for key in obj1.keys():
            assert_approx_objs(obj1[key], obj2[key], rtol=rtol, atol=atol)
    elif isinstance(obj1, list) and isinstance(obj2, list):
        assert len(obj1) == len(obj2)
        for item1, item2 in zip(obj1, obj2):
            assert_approx_objs(item1, item2, rtol=rtol, atol=atol)
    else:
        assert obj1 == pytest.approx(obj2, rel=rtol, abs=atol)


@pytest.fixture(name="instances_dict")
def fx_instances_dict() -> dict:
    with open("test_data/instances_with_errors.json", mode="r", encoding="utf-8") as fp:
        return json.load(fp)


class TestSingleProcess:
    """Tests for single-process inference."""

    @pytest.fixture(scope="class")
    def model(self, model_name: str) -> SpeciesNet:
        return SpeciesNet(model_name)

    def test_predict(self, request, instances_dict, model) -> None:
        predictions_dict1 = model.predict(
            instances_dict=instances_dict, run_mode="single_thread", progress_bars=True
        )
        predictions_dict2 = model.predict(
            instances_dict=instances_dict, run_mode="multi_thread", progress_bars=True
        )
        assert predictions_dict1
        assert predictions_dict2
        assert_approx_objs(predictions_dict1, predictions_dict2, atol=1e-4)
        logging.info("Predictions (%s): %s", request.node.name, predictions_dict1)

    def test_classify(self, request, instances_dict, model) -> None:
        predictions_dict = model.classify(
            instances_dict=instances_dict, run_mode="multi_thread", progress_bars=True
        )
        assert predictions_dict
        logging.info("Classifications (%s): %s", request.node.name, predictions_dict)

    def test_detect(self, request, instances_dict, model) -> None:
        predictions_dict = model.detect(
            instances_dict=instances_dict, run_mode="multi_thread", progress_bars=True
        )
        assert predictions_dict
        logging.info("Detections (%s): %s", request.node.name, predictions_dict)

    def test_ensemble_from_past_runs(self, request, instances_dict, model) -> None:
        predictions_dict = model.ensemble_from_past_runs(
            instances_dict=instances_dict, progress_bars=True
        )
        assert predictions_dict
        logging.info("Ensemble results (%s): %s", request.node.name, predictions_dict)


class TestMultiProcess:
    """Tests for multi-process inference."""

    @pytest.fixture(scope="class")
    def model(self, model_name: str) -> SpeciesNet:
        return SpeciesNet(model_name, multiprocessing=True)

    def test_predict(self, request, instances_dict, model) -> None:
        predictions_dict1 = model.predict(
            instances_dict=instances_dict, run_mode="multi_thread", progress_bars=True
        )
        predictions_dict2 = model.predict(
            instances_dict=instances_dict, run_mode="multi_process", progress_bars=True
        )
        assert predictions_dict1
        assert predictions_dict2
        assert_approx_objs(predictions_dict1, predictions_dict2, atol=1e-4)
        logging.info("Predictions (%s): %s", request.node.name, predictions_dict1)

    def test_batch_predict(self, request, instances_dict, model) -> None:
        predictions_dict1 = model.predict(
            instances_dict=instances_dict, batch_size=1, progress_bars=True
        )
        predictions_dict2 = model.predict(
            instances_dict=instances_dict, batch_size=4, progress_bars=True
        )
        predictions_dict3 = model.predict(
            instances_dict=instances_dict, batch_size=7, progress_bars=True
        )
        assert predictions_dict1
        assert predictions_dict2
        assert predictions_dict3
        assert_approx_objs(predictions_dict1, predictions_dict2, atol=1e-4)
        assert_approx_objs(predictions_dict1, predictions_dict3, atol=1e-4)
        logging.info("Predictions (%s): %s", request.node.name, predictions_dict1)

    def test_classify(self, request, instances_dict, model) -> None:
        predictions_dict1 = model.classify(
            instances_dict=instances_dict, run_mode="multi_thread", progress_bars=True
        )
        predictions_dict2 = model.classify(
            instances_dict=instances_dict, run_mode="multi_process", progress_bars=True
        )
        assert predictions_dict1
        assert predictions_dict2
        assert_approx_objs(predictions_dict1, predictions_dict2, atol=1e-4)
        logging.info("Classifications (%s): %s", request.node.name, predictions_dict1)

    def test_batch_classify(self, request, instances_dict, model) -> None:
        predictions_dict1 = model.classify(
            instances_dict=instances_dict, batch_size=1, progress_bars=True
        )
        predictions_dict2 = model.classify(
            instances_dict=instances_dict, batch_size=4, progress_bars=True
        )
        predictions_dict3 = model.classify(
            instances_dict=instances_dict, batch_size=7, progress_bars=True
        )
        assert predictions_dict1
        assert predictions_dict2
        assert predictions_dict3
        assert_approx_objs(predictions_dict1, predictions_dict2, atol=1e-4)
        assert_approx_objs(predictions_dict1, predictions_dict3, atol=1e-4)
        logging.info("Classifications (%s): %s", request.node.name, predictions_dict1)

    def test_detect(self, request, instances_dict, model) -> None:
        predictions_dict1 = model.detect(
            instances_dict=instances_dict, run_mode="multi_thread", progress_bars=True
        )
        predictions_dict2 = model.detect(
            instances_dict=instances_dict, run_mode="multi_process", progress_bars=True
        )
        assert predictions_dict1
        assert predictions_dict2
        assert_approx_objs(predictions_dict1, predictions_dict2, atol=1e-4)
        logging.info("Detections (%s): %s", request.node.name, predictions_dict1)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to build the geofence release from geofence base with extra manual fixes."""

import copy
import json
from pathlib import Path
from typing import Union

from absl import app
from absl import flags
from absl import logging
import pandas as pd

_BASE = flags.DEFINE_string(
    "base",
    "data/geofence_base.json",
    "Path to the geofence base (JSON). Used as a starting point for constructing the "
    "geofence release.",
)
_FIXES = flags.DEFINE_string(
    "fixes",
    "data/geofence_fixes.csv",
    "Path to the geofence fixes (CSV). Used to correct mistakes in the geofence base.",
)
_TRIM = flags.DEFINE_string(
    "trim",
    None,
    "Path to the labels supported by the model (TXT). Used to trim the geofence "
    "release.",
)
_OUTPUT = flags.DEFINE_string(
    "output",
    None,
    "Output path for writing the geofence release (JSON).",
    required=True,
)

# Handy type alias.
StrPath = Union[str, Path]


def load_geofence_base(path: StrPath) -> dict[str, dict]:

    with open(path, mode="r", encoding="utf-8") as fp:
        data = json.load(fp)
    for label, rules in data.items():
        if label.endswith(";"):
            raise ValueError(
                "Base geofence should provide only species-level rules. "
                f"Found higher taxa rule with the label: `{label}`"
            )
        if "block" in rules:
            raise ValueError("Block rules are not accepted in base geofence.")
    return data


def fix_geofence_base(
    geofence_base: dict[str, dict], fixes_path: StrPath
) -> dict[str, dict]:

    geofence = copy.deepcopy(geofence_base)

    fixes = pd.read_csv(fixes_path, keep_default_na=False)
    for idx, fix in fixes.iterrows():
        label = fix["species"].lower()
        label_parts = label.split(";")
        if len(label_parts) != 5 or not all(label_parts):
            raise ValueError(
                "Fixes should provide only species-level rules. "
                f"Please correct rule #{idx + 1}:\n{fix}"
            )

        rule = fix["rule"].lower()
        if rule not in {"allow", "block"}:
            raise ValueError(
                "Rule types should be either `allow` or `block`. "
                f"Please correct rule #{idx + 1}:\n{fix}"
            )

        country = fix["country_code"]
        state = fix["admin1_region_code"]

        if rule == "allow":
            if label not in geofence:
                continue  # already allowed
            if "allow" not in geofence[label]:
                continue  # already allowed
            if not state:
                geofence[label]["allow"][country] = geofence[label]["allow"].get(
                    country, []
                )
            else:
                curr_country_rule = geofence[label]["allow"].get(country)
                if curr_country_rule is None:  # missing country rule
                    geofence[label]["allow"][country] = [state]
                else:
                    if not curr_country_rule:  # an empty list
                        continue  # already allowed
                    else:  # not an empty list
                        geofence[label]["allow"][country] = sorted(
                            set(curr_country_rule) | {state}
                        )
        else:  # rule == "block"
            if label not in geofence:
                geofence[label] = {"block": {country: [state] if state else []}}
            if "block" not in geofence[label]:
                geofence[label]["block"] = {country: [state] if state else []}
            if not state:
                geofence[label]["block"][country] = geofence[label]["block"].get(
                    country, []
                )
            else:
                curr_country_rule = geofence[label]["block"].get(country)
                if curr_country_rule is None:  # missing country rule
                    geofence[label]["block"][country] = [state]
                else:
                    if not curr_country_rule:  # an empty list
                        continue  # already blocked
                    else:  # not an empty list
                        geofence[label]["block"][country] = sorted(
                            set(curr_country_rule) | {state}
                        )

    return geofence


def propagate_to_higher_taxa(geofence: dict[str, dict]) -> dict[str, dict]:

    new_geofence = {}

    for label, rule in geofence.items():

        label_parts = label.split(";")

        # Keep species rule.
        new_geofence[label] = rule

        # Propagate to higher taxa.
        for taxa_level_end in range(1, 5):
            new_label = ";".join(label_parts[:taxa_level_end]) + (
                ";" * (5 - taxa_level_end)
            )
            if new_label not in new_geofence:
                new_geofence[new_label] = {"allow": {}}

            # Country wide "allow" rules at species level get propagated directly, but
            # regional "allow" rules become country wide "allow" rules at genus level
            # and above.
            if "allow" in rule:
                for country in rule["allow"]:
                    new_geofence[new_label]["allow"][country] = []

    return new_geofence


def trim_to_supported_labels(
    geofence: dict[str, dict], labels_path: StrPath
) -> dict[str, dict]:

    with open(labels_path, mode="r", encoding="utf-8") as fp:
        lines = [line.strip() for line in fp.readlines()]
        labels = set()
        for line in lines:
            label_parts = line.split(";")[1:6]
            for taxa_level_end in range(1, 6):
                new_label = ";".join(label_parts[:taxa_level_end]) + (
                    ";" * (5 - taxa_level_end)
                )
                labels.add(new_label)

    return {k: v for k, v in geofence.items() if k in labels}


def save_geofence(geofence: dict[str, dict], output_path: StrPath) -> None:

    with open(output_path, mode="w", encoding="utf-8") as fp:
        json.dump(geofence, fp, indent=4, sort_keys=True)


def main(argv: list[str]) -> None:
    del argv  # Unused.

    geofence_base = load_geofence_base(_BASE.value)
    geofence_release = fix_geofence_base(geofence_base, _FIXES.value)
    geofence_release = propagate_to_higher_taxa(geofence_release)
    if _TRIM.value:
        logging.info(
            "Trimming to labels (and their corresponding higher taxa) from `%s`.",
            _TRIM.value,
        )
        geofence_release = trim_to_supported_labels(geofence_release, _TRIM.value)
    else:
        logging.info("No trimming was performed.")
    save_geofence(geofence_release, _OUTPUT.value)


if __name__ == "__main__":
    app.run(main)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Useful constants."""

__all__ = [
    "Classification",
    "Detection",
]

import enum
from typing import Optional


class Classification(str, enum.Enum):
    """Enum of common classification values.

    The classifier is not limited to these and can predict other string values as well.
    This enum only contains values with a special meaning during the inference process.
    """

    # pylint: disable=line-too-long
    BLANK = "f1856211-cfb7-4a5b-9158-c0f72fd09ee6;;;;;;blank"
    ANIMAL = "1f689929-883d-4dae-958c-3d57ab5b6c16;;;;;;animal"
    HUMAN = "990ae9dd-7a59-4344-afcb-1b7b21368000;mammalia;primates;hominidae;homo;sapiens;human"
    VEHICLE = "e2895ed5-780b-48f6-8a11-9e27cb594511;;;;;;vehicle"
    UNKNOWN = "f2efdae9-efb8-48fb-8a91-eccf79ab4ffb;no cv result;no cv result;no cv result;no cv result;no cv result;no cv result"


class Detection(str, enum.Enum):
    """Enum of all possible detection values."""

    ANIMAL = "animal"
    HUMAN = "human"
    VEHICLE = "vehicle"

    @classmethod
    def from_category(cls, category: str) -> Optional["Detection"]:
        """Transforms a numeric category from the detector into an enum value.

        Args:
            category: Numeric category from the detector, provided as a string (e.g.
            "1", "2", "3").

        Returns:
            Enum detection value corresponding to the given numeric category. If
            category is not one of "1", "2" or "3", returns `None`.
        """

        category_to_label = {
            "1": Detection.ANIMAL,
            "2": Detection.HUMAN,
            "3": Detection.VEHICLE,
        }
        return category_to_label.get(category)


class Failure(enum.Flag):
    """Enum of flags used to indicate which model components failed during inference."""

    CLASSIFIER = enum.auto()
    DETECTOR = enum.auto()
    GEOLOCATION = enum.auto()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classifier functionality of SpeciesNet.

Defines the SpeciesNetClassifier class, responsible for image classification for
SpeciesNet. It handles loading of classification models, preprocessing of input images,
and generating species predictions.
"""

__all__ = [
    "SpeciesNetClassifier",
]

import time
from typing import Any, Optional

from absl import logging
from humanfriendly import format_timespan
import numpy as np
import PIL.Image
import torch
import torchvision.transforms.functional as F

from speciesnet.constants import Failure
from speciesnet.utils import BBox
from speciesnet.utils import ModelInfo
from speciesnet.utils import PreprocessedImage


class SpeciesNetClassifier:
    """Classifier component of SpeciesNet."""

    IMG_SIZE = 480
    MAX_CROP_RATIO = 0.3
    MAX_CROP_SIZE = 400

    def __init__(
        self, model_name: str, target_species_txt: Optional[str] = None
    ) -> None:
        """Loads the classifier resources.

        Args:
            model_name:
                String value identifying the model to be loaded. It can be a Kaggle
                identifier (starting with `kaggle:`), a HuggingFace identifier (starting
                with `hf:`) or a local folder to load the model from.
        """

        start_time = time.time()

        self.model_info = ModelInfo(model_name)

        # Select the best device available.
        if torch.cuda.is_available():
            self.device = "cuda"
        elif torch.backends.mps.is_available():
            self.device = "mps"
        else:
            self.device = "cpu"

        # Load the model.
        self.model = torch.load(
            self.model_info.classifier, map_location=self.device, weights_only=False
        )

        # Set the model in inference mode.
        self.model.eval()
        for param in self.model.parameters():
            param.requires_grad = False

        # Load the labels.
        with open(self.model_info.classifier_labels, mode="r", encoding="utf-8") as fp:
            self.labels = {idx: line.strip() for idx, line in enumerate(fp.readlines())}

        # Load optional target labels.
        if target_species_txt is not None:
            with open(target_species_txt, mode="r", encoding="utf-8") as fp:
                self.target_labels = [
                    line.strip()
                    for line in fp.readlines()
                    if line.strip() in self.labels.values()
                ]
            labels_to_idx = {label: idx for idx, label in self.labels.items()}
            self.target_idx = [labels_to_idx[label] for label in self.target_labels]

        end_time = time.time()
        logging.info(
            "Loaded SpeciesNetClassifier in %s on %s.",
            format_timespan(end_time - start_time),
            self.device.upper(),
        )

    def preprocess(
        self,
        img: Optional[PIL.Image.Image],
        bboxes: Optional[list[BBox]] = None,
        resize: bool = True,
    ) -> Optional[PreprocessedImage]:
        """Preprocesses an image according to this classifier's needs.

        This method prepares an input image for classification. It handles
        image loading, cropping, and resizing to the expected
        input size for the classifier model.

        In `always_crop` mode images are cropped according to the bounding boxes
        provided. In `full_image` mode the top and bottom of the image are cropped
        to prevent learning correlations between camera brand and species priors.
        See the paper for more details.

        Args:
            img:
                PIL image to preprocess. If `None`, no preprocessing is performed.
            bboxes:
                Optional list of bounding boxes. Needed for some types of classifiers to
                crop the image to specific bounding boxes during preprocessing.
            resize:
                Whether to resize the image to some expected dimensions.

        Returns:
            A preprocessed image, or `None` if no PIL image was provided initially.
        """

        if img is None:
            return None

        img_tensor = F.pil_to_tensor(img)  # HWC to CHW.
        img_tensor = F.convert_image_dtype(img_tensor, torch.float32)

        if self.model_info.type_ == "always_crop":
            # Crop to top bbox if available, otherwise leave image uncropped.
            if bboxes:
                img_tensor = F.crop(
                    img_tensor,
                    int(bboxes[0].ymin * img.height),
                    int(bboxes[0].xmin * img.width),
                    int(bboxes[0].height * img.height),
                    int(bboxes[0].width * img.width),
                )
        elif self.model_info.type_ == "full_image":
            # Crop top and bottom of image.
            target_height = max(
                int(img.height * (1.0 - SpeciesNetClassifier.MAX_CROP_RATIO)),
                img.height - SpeciesNetClassifier.MAX_CROP_SIZE,
            )
            img_tensor = F.center_crop(img_tensor, [target_height, img.width])

        if resize:
            img_tensor = F.resize(
                img_tensor,
                [SpeciesNetClassifier.IMG_SIZE, SpeciesNetClassifier.IMG_SIZE],
                antialias=False,
            )

        img_tensor = F.convert_image_dtype(img_tensor, torch.uint8)
        img_tensor = img_tensor.permute([1, 2, 0])  # CHW to HWC.
        return PreprocessedImage(img_tensor.numpy(), img.width, img.height)

    def predict(
        self, filepath: str, img: Optional[PreprocessedImage]
    ) -> dict[str, Any]:
        """Runs inference on a given preprocessed image.

        Args:
            filepath:
                Location of image to run inference on. Used for reporting purposes only,
                and not for loading the image.
            img:
                Preprocessed image to run inference on. If `None`, a failure message is
                reported back.

        Returns:
            A dict containing either the top-5 classifications for the given image (in
            decreasing order of confidence scores), or a failure message if no
            preprocessed image was provided.
        """

        return self.batch_predict([filepath], [img])[0]

    def batch_predict(
        self, filepaths: list[str], imgs: list[Optional[PreprocessedImage]]
    ) -> list[dict[str, Any]]:
        """Runs inference on a batch of preprocessed images.

        Args:
            filepaths:
                List of image locations to run inference on. Used for reporting purposes
                only, and not for loading the images.
            imgs:
                List of preprocessed images to run inference on. If an image is `None`,
                a corresponding failure message is reported back.

        Returns:
            A list of dict results. Each dict result contains either the top-5
            classifications for the corresponding image (in decreasing order of
            confidence scores), or a failure message if no preprocessed image was
            provided.
        """

        predictions = {}

        inference_filepaths = []
        batch_arr = []
        for filepath, img in zip(filepaths, imgs):
            if img is None:
                predictions[filepath] = {
                    "filepath": filepath,
                    "failures": [Failure.CLASSIFIER.name],
                }
            else:
                inference_filepaths.append(filepath)
                batch_arr.append(img.arr / 255)
        if not batch_arr:
            return list(predictions.values())
        batch_arr = np.stack(batch_arr, axis=0, dtype=np.float32)

        batch_tensor = torch.from_numpy(batch_arr).to(self.device)
        logits = self.model(batch_tensor).cpu()
        scores = torch.softmax(logits, dim=-1)
        scores, indices = torch.topk(scores, k=5, dim=-1)

        for filepath, scores_arr, indices_arr in zip(
            inference_filepaths, scores.numpy(), indices.numpy()
        ):

            predictions[filepath] = {
                "filepath": filepath,
                "classifications": {
                    "classes": [self.labels[idx] for idx in indices_arr],
                    "scores": scores_arr.tolist(),
                },
            }

            if hasattr(self, "target_idx"):
                predictions[filepath]["classifications"].update(
                    {
                        "target_classes": self.target_labels,
                        "target_logits": [
                            float(logits[0][idx]) for idx in self.target_idx
                        ],
                    }
                )

        return [predictions[filepath] for filepath in filepaths]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-module-docstring
# pylint: disable=protected-access
# pylint: disable=redefined-outer-name

import pytest

from speciesnet.geofence_utils import geofence_animal_classification
from speciesnet.geofence_utils import roll_up_labels_to_first_matching_level
from speciesnet.geofence_utils import should_geofence_animal_classification

# fmt: off
# pylint: disable=line-too-long

BLANK = "f1856211-cfb7-4a5b-9158-c0f72fd09ee6;;;;;;blank"
BLANK_FC = ";;;;"
HUMAN = "990ae9dd-7a59-4344-afcb-1b7b21368000;mammalia;primates;hominidae;homo;sapiens;human"
HUMAN_FC = "mammalia;primates;hominidae;homo;sapiens"
VEHICLE = "e2895ed5-780b-48f6-8a11-9e27cb594511;;;;;;vehicle"
VEHICLE_FC = ";;;;"

LION = "ddf59264-185a-4d35-b647-2785792bdf54;mammalia;carnivora;felidae;panthera;leo;lion"
LION_FC = "mammalia;carnivora;felidae;panthera;leo"
PANTHERA_GENUS = "fbb23d07-6677-43db-b650-f99ac452c50f;mammalia;carnivora;felidae;panthera;;panthera species"
PANTHERA_GENUS_FC = "mammalia;carnivora;felidae;panthera;"
FELIDAE_FAMILY = "df8514b0-10a5-411f-8ed6-0f415e8153a3;mammalia;carnivora;felidae;;;cat family"
FELIDAE_FAMILY_FC = "mammalia;carnivora;felidae;;"
CARNIVORA_ORDER = "eeeb5d26-2a47-4d01-a3de-10b33ec0aee4;mammalia;carnivora;;;;carnivorous mammal"
CARNIVORA_ORDER_FC = "mammalia;carnivora;;;"
MAMMALIA_CLASS = "f2d233e3-80e3-433d-9687-e29ecc7a467a;mammalia;;;;;mammal"
MAMMALIA_CLASS_FC = "mammalia;;;;"
ANIMAL_KINGDOM = "1f689929-883d-4dae-958c-3d57ab5b6c16;;;;;;animal"
ANIMAL_KINGDOM_FC = ";;;;"

BROWN_BEAR = "330bb1e9-84d6-4e41-afa9-938aee17ea29;mammalia;carnivora;ursidae;ursus;arctos;brown bear"
BROWN_BEAR_FC = "mammalia;carnivora;ursidae;ursus;arctos"
POLAR_BEAR = "e7f83bf6-df2c-4ce0-97fc-2f233df23ec4;mammalia;carnivora;ursidae;ursus;maritimus;polar bear"
POLAR_BEAR_FC = "mammalia;carnivora;ursidae;ursus;maritimus"
GIANT_PANDA = "85662682-67c1-4ecb-ba05-ba12e2df6b65;mammalia;carnivora;ursidae;ailuropoda;melanoleuca;giant panda"
GIANT_PANDA_FC = "mammalia;carnivora;ursidae;ailuropoda;melanoleuca"
URSUS_GENUS = "5a0f5e3f-c634-4b86-910a-b105cb526a24;mammalia;carnivora;ursidae;ursus;;ursus species"
URSUS_GENUS_FC = "mammalia;carnivora;ursidae;ursus;"
URSIDAE_FAMILY = "ec1a70f4-41c0-4aba-9150-292fb2b7a324;mammalia;carnivora;ursidae;;;bear family"
URSIDAE_FAMILY_FC = "mammalia;carnivora;ursidae;;"

PUMA = "9c564562-9429-405c-8529-04cff7752282;mammalia;carnivora;felidae;puma;concolor;puma"
PUMA_FC = "mammalia;carnivora;felidae;puma;concolor"
SAND_CAT = "e588253d-d61d-4149-a96c-8c245927a80f;mammalia;carnivora;felidae;felis;margarita;sand cat"
SAND_CAT_FC = "mammalia;carnivora;felidae;felis;margarita"

# pylint: enable=line-too-long
# fmt: on


@pytest.fixture
def taxonomy_map():
    return {
        BLANK_FC: BLANK,
        HUMAN_FC: HUMAN,
        VEHICLE_FC: VEHICLE,
        LION_FC: LION,
        PANTHERA_GENUS_FC: PANTHERA_GENUS,
        FELIDAE_FAMILY_FC: FELIDAE_FAMILY,
        CARNIVORA_ORDER_FC: CARNIVORA_ORDER,
        MAMMALIA_CLASS_FC: MAMMALIA_CLASS,
        ANIMAL_KINGDOM_FC: ANIMAL_KINGDOM,
        BROWN_BEAR_FC: BROWN_BEAR,
        POLAR_BEAR_FC: POLAR_BEAR,
        GIANT_PANDA_FC: GIANT_PANDA,
        URSUS_GENUS_FC: URSUS_GENUS,
        URSIDAE_FAMILY_FC: URSIDAE_FAMILY,
    }


@pytest.fixture
def geofence_map():
    return {
        LION_FC: {
            "allow": {
                "KEN": [],
                "TZA": [],
            }
        },
        PANTHERA_GENUS_FC: {
            "allow": {
                "KEN": [],
                "TZA": [],
                "USA": ["AK", "CA"],
            }
        },
        FELIDAE_FAMILY_FC: {
            "allow": {
                "FRA": [],
                "KEN": [],
                "TZA": [],
                "USA": [],
            },
            "block": {
                "FRA": [],
                "USA": ["NY"],
            },
        },
        SAND_CAT_FC: {
            "block": {
                "AUS": [],
            },
        },
        URSIDAE_FAMILY_FC: {
            "block": {
                "GBR": [],
            },
        },
    }


class TestGeofenceUtils:
    """Tests for the geofence utility functions."""

    def test_should_geofence_animal_classification(self, geofence_map) -> None:

        # Test when country is not provided.
        assert not should_geofence_animal_classification(
            LION,
            country=None,
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )

        # Test when label is not in the geofence map.
        assert not should_geofence_animal_classification(
            PUMA,
            country="USA",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert not should_geofence_animal_classification(
            PUMA,
            country="USA",
            admin1_region="CA",
            geofence_map=geofence_map,
            enable_geofence=True,
        )

        # Test "allow" rules from the geofence map.
        assert should_geofence_animal_classification(
            LION,
            country="GBR",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert not should_geofence_animal_classification(
            LION,
            country="KEN",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert not should_geofence_animal_classification(
            PANTHERA_GENUS,
            country="USA",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert should_geofence_animal_classification(
            PANTHERA_GENUS,
            country="USA",
            admin1_region="NY",
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert not should_geofence_animal_classification(
            PANTHERA_GENUS,
            country="USA",
            admin1_region="CA",
            geofence_map=geofence_map,
            enable_geofence=True,
        )

        # Test "block" rules from the geofence map.
        assert should_geofence_animal_classification(
            FELIDAE_FAMILY,
            country="FRA",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert not should_geofence_animal_classification(
            FELIDAE_FAMILY,
            country="TZA",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert not should_geofence_animal_classification(
            FELIDAE_FAMILY,
            country="USA",
            admin1_region="CA",
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert should_geofence_animal_classification(
            FELIDAE_FAMILY,
            country="USA",
            admin1_region="NY",
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert not should_geofence_animal_classification(
            SAND_CAT,
            country="GBR",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert should_geofence_animal_classification(
            SAND_CAT,
            country="AUS",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=True,
        )

    def test_should_geofence_animal_classification_disabled(self, geofence_map) -> None:

        # Test when country is not provided.
        assert not should_geofence_animal_classification(
            LION,
            country=None,
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )

        # Test when label is not in the geofence map.
        assert not should_geofence_animal_classification(
            PUMA,
            country="USA",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            PUMA,
            country="USA",
            admin1_region="CA",
            geofence_map=geofence_map,
            enable_geofence=False,
        )

        # Test "allow" rules from the geofence map.
        assert not should_geofence_animal_classification(
            LION,
            country="GBR",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            LION,
            country="KEN",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            PANTHERA_GENUS,
            country="USA",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            PANTHERA_GENUS,
            country="USA",
            admin1_region="NY",
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            PANTHERA_GENUS,
            country="USA",
            admin1_region="CA",
            geofence_map=geofence_map,
            enable_geofence=False,
        )

        # Test "block" rules from the geofence map.
        assert not should_geofence_animal_classification(
            FELIDAE_FAMILY,
            country="FRA",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            FELIDAE_FAMILY,
            country="TZA",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            FELIDAE_FAMILY,
            country="USA",
            admin1_region="CA",
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            FELIDAE_FAMILY,
            country="USA",
            admin1_region="NY",
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            SAND_CAT,
            country="GBR",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )
        assert not should_geofence_animal_classification(
            SAND_CAT,
            country="AUS",
            admin1_region=None,
            geofence_map=geofence_map,
            enable_geofence=False,
        )

    def test_roll_up_labels_to_first_matching_level(
        self, taxonomy_map, geofence_map
    ) -> None:
        # pylint: disable=unnecessary-lambda-assignment

        predictions = [
            BROWN_BEAR,
            POLAR_BEAR,
            GIANT_PANDA,
            BLANK,
            LION,
            HUMAN,
            ANIMAL_KINGDOM,
        ]

        # Test rollups to species level.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["species"],
            non_blank_threshold=0.9,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0]) is None
        assert rollup_fn([0.9, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0]) is None
        assert rollup_fn([1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]) == (
            BROWN_BEAR,
            pytest.approx(1.0),
            "classifier+rollup_to_species",
        )

        # Test rollups to genus level.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["genus"],
            non_blank_threshold=0.9,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.6, 0.2, 0.01, 0.01, 0.01, 0.01, 0.01]) is None
        assert rollup_fn([0.7, 0.25, 0.01, 0.01, 0.01, 0.01, 0.01]) == (
            URSUS_GENUS,
            pytest.approx(0.95),
            "classifier+rollup_to_genus",
        )

        # Test rollups to family level.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["family"],
            non_blank_threshold=0.8,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.4, 0.1, 0.1, 0.1, 0.1, 0.0, 0.0]) is None
        assert rollup_fn([0.4, 0.21, 0.2, 0.0, 0.0, 0.0, 0.0]) == (
            URSIDAE_FAMILY,
            pytest.approx(0.81),
            "classifier+rollup_to_family",
        )

        # Test rollups to order level.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["order"],
            non_blank_threshold=0.8,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.3, 0.2, 0.1, 0.1, 0.1, 0.0, 0.0]) is None
        assert rollup_fn([0.3, 0.2, 0.1, 0.1, 0.23, 0.0, 0.0]) == (
            CARNIVORA_ORDER,
            pytest.approx(0.83),
            "classifier+rollup_to_order",
        )

        # Test rollups to class level.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["class"],
            non_blank_threshold=0.8,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.2, 0.2, 0.1, 0.1, 0.1, 0.1, 0.0]) is None
        assert rollup_fn([0.2, 0.2, 0.1, 0.1, 0.22, 0.1, 0.0]) == (
            MAMMALIA_CLASS,
            pytest.approx(0.82),
            "classifier+rollup_to_class",
        )

        # Test rollups to kingdom level.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["kingdom"],
            non_blank_threshold=0.81,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.2, 0.2, 0.1, 0.1, 0.1, 0.1, 0.1]) is None
        assert rollup_fn([0.2, 0.2, 0.1, 0.1, 0.23, 0.1, 0.1]) == (
            ANIMAL_KINGDOM,
            pytest.approx(0.93),
            "classifier+rollup_to_kingdom",
        )

        # Test rollups when multiple taxonomy levels are specified.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["genus", "family", "order", "class", "kingdom"],
            non_blank_threshold=0.75,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.6, 0.1, 0.1, 0.1, 0.1, 0.0, 0.0]) == (
            URSIDAE_FAMILY,
            pytest.approx(0.8),
            "classifier+rollup_to_family",
        )

        # Test rollups when multiple score sums pass the non blank threshold.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["species"],
            non_blank_threshold=0.1,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.2, 0.3, 0.15, 0.0, 0.35, 0.0, 0.0]) == (
            LION,
            pytest.approx(0.35),
            "classifier+rollup_to_species",
        )

        # Test rollups when the BLANK score dominates all the others.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country=None,
            admin1_region=None,
            target_taxonomy_levels=["species", "genus", "family", "order", "class"],
            non_blank_threshold=0.4,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.1, 0.2, 0.2, 0.45, 0.0, 0.0, 0.0]) == (
            URSIDAE_FAMILY,
            pytest.approx(0.5),
            "classifier+rollup_to_family",
        )

        # Test rollups with geofencing.
        rollup_fn = lambda scores: roll_up_labels_to_first_matching_level(
            labels=predictions,
            scores=scores,
            country="GBR",
            admin1_region=None,
            target_taxonomy_levels=["species", "genus", "family", "order", "class"],
            non_blank_threshold=0.4,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert rollup_fn([0.1, 0.2, 0.2, 0.45, 0.0, 0.0, 0.0]) == (
            CARNIVORA_ORDER,
            pytest.approx(0.5),
            "classifier+rollup_to_order",
        )

        # Test rollups to invalid levels.
        with pytest.raises(ValueError):
            roll_up_labels_to_first_matching_level(
                labels=predictions,
                scores=[0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
                country=None,
                admin1_region=None,
                target_taxonomy_levels=["invalid_level"],
                non_blank_threshold=0.3,
                taxonomy_map=taxonomy_map,
                geofence_map=geofence_map,
                enable_geofence=True,
            )

    def test_geofence_animal_classification(self, taxonomy_map, geofence_map) -> None:
        # pylint: disable=unnecessary-lambda-assignment

        predictions = [
            LION,
            POLAR_BEAR,
            BLANK,
            FELIDAE_FAMILY,
        ]

        # Test when no geofencing is needed.
        geofence_fn = lambda scores: geofence_animal_classification(
            labels=predictions,
            scores=scores,
            country="TZA",
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert geofence_fn([0.4, 0.3, 0.2, 0.1]) == (
            LION,
            pytest.approx(0.4),
            "classifier",
        )

        # Test with geofencing and rollup to family level or above.
        geofence_fn = lambda scores: geofence_animal_classification(
            labels=predictions,
            scores=scores,
            country="USA",
            admin1_region=None,
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert geofence_fn([0.4, 0.3, 0.2, 0.1]) == (
            FELIDAE_FAMILY,
            pytest.approx(0.5),
            "classifier+geofence+rollup_to_family",
        )
        geofence_fn = lambda scores: geofence_animal_classification(
            labels=predictions,
            scores=scores,
            country="USA",
            admin1_region="NY",
            taxonomy_map=taxonomy_map,
            geofence_map=geofence_map,
            enable_geofence=True,
        )
        assert geofence_fn([0.4, 0.3, 0.2, 0.1]) == (
            CARNIVORA_ORDER,
            pytest.approx(0.8),
            "classifier+geofence+rollup_to_order",
        )
